/*
 * host_utils.c -- utils for host
 *
 * Author: Sukru Senli sukru.senli@inteno.se
 * Author: Nevadita Chatterjee nevadita.chatterjee@iopsys.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <libubus.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>

#include <netlink/netlink.h>
#include <netlink/utils.h>

#include <netlink/route/rtnl.h>
#include <netlink/route/neighbour.h>
#include <netlink/route/addr.h>
#include <netlink/route/link.h>
#include <netlink/genl/ctrl.h>
#include <netlink/genl/genl.h>
#include <netlink/attr.h>

#include <uci.h>
#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host.h"
#include "host_utils.h"
#include "host_nodes.h"

#ifndef TLV_TYPE_END_OF_MESSAGE
#define TLV_TYPE_END_OF_MESSAGE 0x00
#endif

struct in_addr src;
struct in_addr dst;
struct sockaddr_ll me;
struct sockaddr_ll he;
int sock_fd;

typedef struct nl_req_s nl_req_t;

struct nl_req_s {
	struct nlmsghdr hdr;
	struct rtgenmsg gen;
};

struct topo_nlevent {
	struct uloop_fd uloop;
	void (*error_cb)(struct topo_nlevent *e, int error);
	void (*event_cb)(struct topo_nlevent *e);
};

struct event_socket {
	struct topo_nlevent ev;
	struct nl_sock *sock;
	int sock_bufsize;
};

static void handle_error(struct topo_nlevent *e, int error)
{
	struct event_socket *ev_sock = container_of(e, struct event_socket, ev);

	if (error != ENOBUFS)
		goto err;

	ev_sock->sock_bufsize *= 2;
	if (nl_socket_set_buffer_size(ev_sock->sock, ev_sock->sock_bufsize, 0))
		goto err;

	return;

err:
	e->uloop.cb = NULL;
	uloop_fd_delete(&e->uloop);
}

static void recv_nlevents(struct topo_nlevent *e)
{
	struct event_socket *ev_sock = container_of(e, struct event_socket, ev);

	nl_recvmsgs_default(ev_sock->sock);
}


static struct event_socket rtnl_event = {
	.ev = {
		.uloop = {.fd = -1, },
		.error_cb = handle_error,
		.event_cb = recv_nlevents,
	},
	.sock = NULL,
	.sock_bufsize = 0x20000,
};

static int nlevents_cb(struct nl_msg *msg, void *arg);
static int handle_nlevents_neigh(struct topologyd_private *priv,
		struct nlmsghdr *hdr, bool add);
static int handle_neigh_tbl_change(struct topologyd_private *priv, bool add,
		char *ifname, uint8_t *macaddr,
		struct ip_address *ip, uint16_t state);
static int host_part_of_zone(struct host_config *cfg, uint32_t ifindex);

int host_get_netlink_ip(uint8_t *mac_addr, char *ipv4_str, char *device)
{
	struct rtnl_neigh *neigh;
	struct nl_object *nobj;
	struct nl_cache *cache;
	uint32_t ifindex = 0;
	struct nl_sock *sk;
	int i, num;
	int ret;
	char *ifname = NULL;

	dbg("Inside %s %d\n", __func__, __LINE__);
	sk = nl_socket_alloc();
	if (!sk) {
		err("Unable to open nl event socket\n");
		return -1;
	}

	if (nl_connect(sk, NETLINK_ROUTE) < 0) {
		nl_socket_free(sk);
		return -1;
	}

	ret = rtnl_neigh_alloc_cache(sk, &cache);
	if (ret) {
		nl_socket_free(sk);
		return -1;
	}


	num = nl_cache_nitems(cache);
	nobj = nl_cache_get_first(cache);
	neigh = (struct rtnl_neigh *)nobj;

	for (i = 0; i < num; i++) {
		struct nl_addr *lladdr;
		struct nl_addr *ipaddr;
		struct ip_address ip = {0};
		uint8_t hwaddr[6] = {0};
		char addr_str[24];

		nl_object_get((struct nl_object *) neigh);

		ifindex = rtnl_neigh_get_ifindex(neigh);
		lladdr = rtnl_neigh_get_lladdr(neigh);
		if (lladdr)
			memcpy(hwaddr, nl_addr_get_binary_addr(lladdr),
				nl_addr_get_len(lladdr));

		if (hwaddr_is_zero(hwaddr)) {
			nl_object_put((struct nl_object *) neigh);
			nobj = nl_cache_get_next(nobj);
			neigh = (struct rtnl_neigh *)nobj;
			continue;
		}

		ipaddr = rtnl_neigh_get_dst(neigh);
		if (ipaddr) {
			ip.family = nl_addr_get_family(ipaddr);
			if (ip.family == AF_INET6) {
				nl_object_put((struct nl_object *) neigh);
				nobj = nl_cache_get_next(nobj);
				neigh = (struct rtnl_neigh *)nobj;
				continue;
			}
			if (ip.family == AF_INET) {
				memcpy(&ip.addr, nl_addr_get_binary_addr(ipaddr),
				       nl_addr_get_len(ipaddr));
			}
		}

		char *addr_s = nl_addr2str(ipaddr, addr_str, sizeof(addr_str));

		if (addr_s == NULL) {
			/* error */
			err("nl_addr2str failed\n");
			nl_object_put((struct nl_object *) neigh);
			nl_cache_free(cache);
			nl_socket_free(sk);
			return -1;
		}

		ret = memcmp(hwaddr, mac_addr, 6);
		if (ret == 0) {
			strncpy(ipv4_str, addr_str, 24);
			ifname = if_indextoname(ifindex, device);
			if (ifname == NULL) {
				nl_object_put((struct nl_object *) neigh);
				nl_cache_free(cache);
				nl_socket_free(sk);
				return -1;
			}
			dbg("intfname = [%s] ipv4addr= [%s] mac=["MACFMT"]\n",
			    device, ipv4_str, MAC2STR(hwaddr));

			nl_object_put((struct nl_object *) neigh);
			break;
		}

		nl_object_put((struct nl_object *) neigh);
		nobj = nl_cache_get_next(nobj);
		neigh = (struct rtnl_neigh *)nobj;
	}

	nl_cache_free(cache);
	nl_socket_free(sk);

	return 0;
}

int host_get_netlink_ip6(uint8_t *mac_addr, struct host_node *p)
{
	struct rtnl_neigh *neigh;
	struct nl_object *nobj;
	struct nl_cache *cache;
	//uint32_t ifindex = 0;
	struct nl_sock *sk;
	int i, num;
	int ret;

	dbg("Inside %s %d\n", __func__, __LINE__);
	sk = nl_socket_alloc();
	if (!sk) {
		err("Unable to open nl event socket\n");
		return -1;
	}

	if (nl_connect(sk, NETLINK_ROUTE) < 0) {
		nl_socket_free(sk);
		return -1;
	}

	ret = rtnl_neigh_alloc_cache(sk, &cache);
	if (ret) {
		nl_socket_free(sk);
		return -1;
	}

	num = nl_cache_nitems(cache);
	nobj = nl_cache_get_first(cache);
	neigh = (struct rtnl_neigh *)nobj;
	p->ipv6addr_count = 0;

	for (i = 0; i < num; i++) {
		struct nl_addr *lladdr;
		struct nl_addr *ipaddr;
		struct ip_address ip = {0};
		uint8_t hwaddr[6] = {0};
		char addr_str[128];

		nl_object_get((struct nl_object *) neigh);

		//ifindex = rtnl_neigh_get_ifindex(neigh);
		lladdr = rtnl_neigh_get_lladdr(neigh);
		if (lladdr)
			memcpy(hwaddr, nl_addr_get_binary_addr(lladdr),
				nl_addr_get_len(lladdr));

		if (hwaddr_is_zero(hwaddr)) {
			nl_object_put((struct nl_object *) neigh);
			nobj = nl_cache_get_next(nobj);
			neigh = (struct rtnl_neigh *)nobj;
			continue;
		}

		ipaddr = rtnl_neigh_get_dst(neigh);
		if (ipaddr) {
			ip.family = nl_addr_get_family(ipaddr);
			if (ip.family == AF_INET6) {
				memcpy(&ip.addr, nl_addr_get_binary_addr(ipaddr),
					nl_addr_get_len(ipaddr));
			} else {
				nl_object_put((struct nl_object *) neigh);
				nobj = nl_cache_get_next(nobj);
				neigh = (struct rtnl_neigh *)nobj;
				continue;
			}
		}
		char *addr_s = nl_addr2str(ipaddr, addr_str, sizeof(addr_str));

		if (addr_s == NULL) {
			/* error */
			err("nl_addr2str failed\n");
			nl_object_put((struct nl_object *) neigh);
			nl_cache_free(cache);
			nl_socket_free(sk);
			return -1;
		}

		ret = memcmp(hwaddr, mac_addr, 6);
		if (ret == 0) {
			if (p->ipv6addr_count >= HOST_MAX_IPV6ADDR) {
				nl_object_put((struct nl_object *) neigh);
				break;
			}
			memset(p->ipv6addr[p->ipv6addr_count], 0, 128);
			strncpy(p->ipv6addr[p->ipv6addr_count], addr_str, 127);
			dbg("ipv6addr= [%s] mac=["MACFMT"]\n", p->ipv6addr[p->ipv6addr_count], MAC2STR(hwaddr));
			p->ipv6addr_count = p->ipv6addr_count + 1;
		}

		nl_object_put((struct nl_object *) neigh);
		nobj = nl_cache_get_next(nobj);
		neigh = (struct rtnl_neigh *)nobj;
	}

	nl_cache_free(cache);
	nl_socket_free(sk);

	return 0;
}

void _E1B(uint8_t **packet_ppointer, uint8_t *memory_pointer)
{
	*memory_pointer = **packet_ppointer;
	(*packet_ppointer) += 1;
}

// Extract/insert 2 bytes
void _E2B(uint8_t **packet_ppointer, uint16_t *memory_pointer)
{
	uint16_t mem = 0;
	*(((uint8_t *) &mem) + 0) = **packet_ppointer;
	(*packet_ppointer)++;
	*(((uint8_t *) &mem) + 1) = **packet_ppointer;
	(*packet_ppointer)++;
	*memory_pointer = ntohs(mem);
}

// Extract/insert N bytes (ignore endianess)
void _EnB(uint8_t **packet_ppointer,
		void *memory_pointer,
		uint32_t n)
{
	memcpy(memory_pointer, *packet_ppointer, n);
	(*packet_ppointer) += n;
}

// count includes end-of-message tlv
uint8_t count_tlv_present_instream_unsafe(uint8_t *tlv_stream)
{
	uint8_t *p = tlv_stream;
	uint8_t tlv_type, count = 0;
	uint16_t tlv_len = 0;

	if (tlv_stream == NULL) {
		err("tlv Stream is empty\n");
		return 0;
	}

	do {
		_E1B(&p, &tlv_type);
		_E2B(&p, &tlv_len);

		p = p + tlv_len;
		count++;
	} while (tlv_type != TLV_TYPE_END_OF_MESSAGE);

	return count;
}

uint16_t get_tlv_packet_len(uint8_t *tlv_stream)
{
	uint8_t *p;
	uint8_t type;
	uint16_t len;

	if (tlv_stream == NULL)
		return 0;

	p = tlv_stream;
	_E1B(&p, &type);
	_E2B(&p, &len);

	// tlv packet does not have length of tlv_type(1 Byte) and
	// tlv_len(2 Bytes) included
	return len + 3;
}

uint8_t get_tlv_packet_type(uint8_t *tlv_stream)
{
	uint8_t *p;
	uint8_t type;
	uint16_t len;

	if (tlv_stream == NULL)
		return 0;

	p = tlv_stream;
	_E1B(&p, &type);
	_E2B(&p, &len);

	return type;
}

static void receive_nlevents(struct uloop_fd *u, unsigned int events)
{
	struct topo_nlevent *e = container_of(u, struct topo_nlevent, uloop);

	if (u->error) {
		int ret = -1;
		socklen_t ret_len = sizeof(ret);

		u->error = false;
		if (e->error_cb &&
			getsockopt(u->fd, SOL_SOCKET, SO_ERROR, &ret, &ret_len) == 0) {
			e->error_cb(e, ret);
		}
	}

	if (e->event_cb) {
		e->event_cb(e);
		return;
	}
}

int register_nlevents(struct topologyd_private *priv)
{
	struct nl_sock *sk;

	sk = nl_socket_alloc();
	if (!sk) {
		err("Unable to open nl event socket: %m");
		return -1;
	}

	if (nl_connect(sk, NETLINK_ROUTE) < 0) {
		nl_socket_free(sk);
		return -1;
	}

	rtnl_event.sock = sk;

	if (nl_socket_set_buffer_size(rtnl_event.sock, rtnl_event.sock_bufsize, 0)) {
		err("%s: %d\n", __func__, __LINE__);
		goto out_err;
	}

	nl_socket_disable_seq_check(rtnl_event.sock);

	nl_socket_modify_cb(rtnl_event.sock, NL_CB_VALID, NL_CB_CUSTOM,
		nlevents_cb, priv);

	if (nl_socket_add_memberships(rtnl_event.sock,
		RTNLGRP_NEIGH, RTNLGRP_LINK, 0))
		goto out_err;

	rtnl_event.ev.uloop.fd = nl_socket_get_fd(rtnl_event.sock);
	rtnl_event.ev.uloop.cb = receive_nlevents;
	uloop_fd_add(&rtnl_event.ev.uloop, ULOOP_READ |
		((rtnl_event.ev.error_cb) ? ULOOP_ERROR_CB : 0));

	return 0;

out_err:
	if (rtnl_event.sock) {
		nl_socket_free(rtnl_event.sock);
		rtnl_event.sock = NULL;
		rtnl_event.ev.uloop.fd = -1;
	}

	return -1;
}

void unregister_nlevents(void)
{
	if (rtnl_event.sock) {
		uloop_fd_delete(&rtnl_event.ev.uloop);
		rtnl_event.ev.uloop.fd = -1;
		nl_socket_free(rtnl_event.sock);
		rtnl_event.sock = NULL;
	}
}

static int nlevents_cb(struct nl_msg *msg, void *arg)
{
	struct nlmsghdr *hdr = nlmsg_hdr(msg);
	struct topologyd_private *priv = arg;
	int ret = NL_SKIP;
	bool add = false;


	switch (hdr->nlmsg_type) {
	case RTM_NEWLINK:
		add = true;
	case RTM_DELLINK:
		break;
#if 0
	//TODO: when needed
	case RTM_NEWADDR:
		add = true;
	case RTM_DELADDR:
		ret = i1905_handle_nlevents_addr(priv, hdr, add);
		break;
#endif
	case RTM_NEWNEIGH:
		add = true;
	case RTM_DELNEIGH:
		ret = handle_nlevents_neigh(priv, hdr, add);
		break;
	default:
		break;
	}

	return ret;
}

static int handle_nlevents_neigh(struct topologyd_private *priv,
		struct nlmsghdr *hdr, bool add)
{
	struct ndmsg *ndm = nlmsg_data(hdr);
	struct nlattr *nla[__NDA_MAX];
	uint8_t macaddr[6] = {0};
	struct ip_address ip;
	char ipbuf[256] = {0};
	char ifname[16] = {0};
	char state[128] = {0};
	int ret = 0;


	if (!nlmsg_valid_hdr(hdr, sizeof(*ndm)))
		return NL_SKIP;

	nlmsg_parse(hdr, sizeof(*ndm), nla, __NDA_MAX - 1, NULL);
	if (!nla[NDA_DST])
		return NL_SKIP;

	nla_memcpy(&ip.addr, nla[NDA_DST], sizeof(ip.addr));
	nla_memcpy(macaddr, nla[NDA_LLADDR], sizeof(macaddr));
	ip.family = ndm->ndm_family;

	if (IN6_IS_ADDR_LINKLOCAL(&ip.addr) || IN6_IS_ADDR_MULTICAST(&ip.addr))
		return NL_SKIP;

	if (ndm->ndm_family == AF_INET || ndm->ndm_family == AF_INET6)
		inet_ntop(ip.family, &ip.addr, ipbuf, sizeof(ipbuf));

	if_indextoname(ndm->ndm_ifindex, ifname);
	dbg("Netlink neigh %s on %s   state = %s\n", ipbuf, ifname,
		rtnl_neigh_state2str(ndm->ndm_state, state, sizeof(state)));

	ret = host_part_of_zone(&priv->host_global_cfg, ndm->ndm_ifindex);
	if (ret != 0)
		return NL_OK;

	handle_neigh_tbl_change(priv, add, ifname, macaddr, &ip,
		ndm->ndm_state);

	return NL_OK;
}

static int handle_neigh_tbl_change(struct topologyd_private *priv, bool add,
		char *ifname, uint8_t *macaddr,
		struct ip_address *ip,
		uint16_t state)
{
	struct host_node *hn = NULL;

	if (hwaddr_is_zero(macaddr))
		return 0;

	hn = host_node_lookup(priv->host.node_htable, macaddr);
	if (!hn) {
		hn = host_topo_node_add(priv, NULL, macaddr, 0);
		if (!hn) {
			err("Failed to add non1905 node in the hosts\n");
			return -1;
		}
		if (hn->is_ipaddr)
			host_send_client_event(priv, hn, 1);
	}
	return 0;

}

int host_get_neighbors(struct topologyd_private *priv)
{
	struct rtnl_neigh *neigh;
	struct nl_object *nobj;
	struct nl_cache *cache;
	struct nl_sock *sk;
	int i, num;
	int ret;
	struct host_node *hn = NULL;
	uint32_t ifindex = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);
	sk = nl_socket_alloc();
	if (!sk) {
		err("Unable to open nl event socket\n");
		return -1;
	}

	if (nl_connect(sk, NETLINK_ROUTE) < 0) {
		nl_socket_free(sk);
		return -1;
	}

	ret = rtnl_neigh_alloc_cache(sk, &cache);
	if (ret) {
		nl_socket_free(sk);
		return -1;
	}
	num = nl_cache_nitems(cache);
	nobj = nl_cache_get_first(cache);
	neigh = (struct rtnl_neigh *)nobj;

	for (i = 0; i < num; i++) {
		struct nl_addr *lladdr;
		uint8_t hwaddr[6] = {0};

		nl_object_get((struct nl_object *) neigh);

		ifindex = rtnl_neigh_get_ifindex(neigh);
		lladdr = rtnl_neigh_get_lladdr(neigh);
		if (lladdr)
			memcpy(hwaddr, nl_addr_get_binary_addr(lladdr),
				nl_addr_get_len(lladdr));

		if (hwaddr_is_zero(hwaddr)) {
			nl_object_put((struct nl_object *) neigh);
			nobj = nl_cache_get_next(nobj);
			neigh = (struct rtnl_neigh *)nobj;
			continue;
		}

		ret = host_part_of_zone(&priv->host_global_cfg, ifindex);
		if (ret != 0) {
			nl_object_put((struct nl_object *) neigh);
			nobj = nl_cache_get_next(nobj);
			neigh = (struct rtnl_neigh *)nobj;
			continue;
		}

		/* Here we need to put this macaddress in the host table*/
		hn = host_node_lookup(priv->host.node_htable, hwaddr);
		if (!hn) {
			dbg("adding mac address"MACFMT"\n", MAC2STR(hwaddr));
			hn = host_topo_node_add(priv, NULL, hwaddr, 0);
			if (!hn) {
				err("Failed to add non1905 node in the hosts\n");
				nl_object_put((struct nl_object *) neigh);
				nobj = nl_cache_get_next(nobj);
				neigh = (struct rtnl_neigh *)nobj;
				continue;
			}
			if (hn->is_ipaddr)
				host_send_client_event(priv, hn, 1);
		}

		nl_object_put((struct nl_object *) neigh);
		nobj = nl_cache_get_next(nobj);
		neigh = (struct rtnl_neigh *)nobj;
	}

	nl_cache_free(cache);
	nl_socket_free(sk);

	return 0;
}

static int host_part_of_zone(struct host_config *cfg, uint32_t ifindex)
{
	uint32_t zoneindex;
	int i  = 0;

	dbg("Inside %s %d ifindex is %d\n", __func__, __LINE__, ifindex);
	for (i = 0; i < cfg->num_ifname; i++) {
		zoneindex = if_nametoindex((char *)cfg->ifname[i]);
		if (zoneindex == ifindex) {
			return 0;
		}
	}
	return -1;
}
