/*
 * topologyd.c - topology for multiap network ubus interface
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdbool.h>
#include <pthread.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <uci.h>
#include <string.h>

#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "nodes.h"
#include "config.h"
#include "json_utils.h"
#include "topo_ieee1905.h"
#include "host_nodes.h"
#include "host.h"
#include "cmdu_validate.h"
#include "ieee1905_deps.h"
#include "host_ieee1905_utils.h"
#include "host_utils.h"
#include "host_wan_stats.h"

extern const char *ubus_socket;
uint32_t topology_log_max;
static bool g_registrar_freq_band[3];


static void topology_send_query_response(struct topologyd_private *priv,
	uint16_t msg_type, uint8_t *origin, uint8_t *source, uint16_t msg_mid,
	char *ifname);

void topologyd_process_topology_notification(struct cmdu_buff *cstruct,
						struct topologyd_private *priv);


static void ieee1905_info_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct topologyd_private *p;
	struct blob_attr *cur;
	int rem;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy attr[2] = {
		[0] = { .name = "ieee1905id", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "registrar_band", .type = BLOBMSG_TYPE_ARRAY },
	};

	if (!msg || !req->priv)
		return;

	p = (struct topologyd_private *)req->priv;

	blobmsg_parse(attr, 2, tb, blob_data(msg), blob_len(msg));

	if (tb[0]) {
		char *mac_str = NULL;
		mac_str = blobmsg_get_string(tb[0]);
		if (mac_str) {
			dbg(" IEE1905 AL mac is %s\n", mac_str);
			hwaddr_aton(mac_str, p->ieee1905_macaddr);
			hwaddr_aton(mac_str, p->selfnode.hwaddr);
		}
	}
	if (tb[1]) {
		blobmsg_for_each_attr(cur, tb[1], rem) {
			char *freq_band = strdup(blobmsg_get_string(cur));

			if (!freq_band)
				return;

			dbg(" IEE1905 freq band is %s\n", freq_band);
			if ((strcasestr(freq_band, "2.4GHz")) != NULL)
				g_registrar_freq_band[0] = 1;
			if ((strcasestr(freq_band, "5GHz")) != NULL)
				g_registrar_freq_band[1] = 1;
			if ((strcasestr(freq_band, "60GHz")) != NULL)
				g_registrar_freq_band[2] = 1;
			free(freq_band);
		}
	}
}

int ieee1905_msg_from_ubus(struct blob_attr *msg, uint8_t *source,
			   uint8_t *origin, uint16_t *msg_type,
			   uint16_t *msg_mid, uint8_t **tlv,
			   uint32_t *b_len, char *ingress_if_name)
{
	json_object *jroot = NULL;
	char *str;
	char *tlv_str;
	char *if_name, *origin_str;
	char *mac_str;
	int len;

	if (!msg)
		return 1;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		*msg_type = _json_obj_get_int(jroot, "type");
		*msg_mid = _json_obj_get_int(jroot, "mid");
		origin_str = _json_obj_get_string(jroot, "origin");
		if (origin_str) {
			hwaddr_aton(origin_str, origin);
			free(origin_str);
		}
		mac_str = _json_obj_get_string(jroot, "source");
		if (mac_str) {
			hwaddr_aton(mac_str, source);
			free(mac_str);
		}
		if_name = _json_obj_get_string(jroot, "ifname");
		if (if_name != NULL) {
			strncpy(ingress_if_name, if_name, strlen(if_name));
			free(if_name);
		}

		tlv_str = _json_obj_get_string(jroot, "cmdu");
		if (tlv_str == NULL) {
			json_safe_put(jroot);
			return 1;
		}

		len = strlen(tlv_str) - 16;
	//	dbg(" The TLV string is %s LENGTH Id %d", tlv_str, len);
		*b_len = (len/2);
	//	dbg("The binary length of the tlv is %d", *b_len);
		*tlv = (uint8_t *) malloc((*b_len) * sizeof(uint8_t));
		if (*tlv == NULL) {
			err("No Memory\n");
			free(tlv_str);
			json_safe_put(jroot);
			return 1;
		}

		strtob(tlv_str+16, *b_len, *tlv);
		free(tlv_str);
		json_safe_put(jroot);
	}

	return 0;
}

int topologyd_get_ieee1905_macaddr(struct topologyd_private *p)
{
	struct blob_buf bb = { 0 };
	int ret = 0;
	uint32_t id = 0;

	blob_buf_init(&bb, 0);
	ubus_lookup_id(p->ctx, I1905_OBJ, &id);
	ret = ubus_invoke(p->ctx, id, "info", bb.head,
				ieee1905_info_response_cb, p, 1000);
	if (ret)
		err("Failed to get ieee1905 info\n");

	blob_buf_free(&bb);

	return ret;
}

int topologyd_get_albridge_macaddr(struct topologyd_private *p)
{
	char ifname[MAX_IFNAME * 16] = { 0 };
	int ret = 0, i = 0;
	int num_ifname = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	/* call host_config.c to get ifname*/
	if (!config_get_al_ifname((char *)ifname, &num_ifname))
		 return 1;

	/*
	call libeasy to get hwaddr of ifname
	store hwaddr in p->aliface_macaddr
	*/
	for (i = 0; i < num_ifname; i++) {
		uint8_t macaddr[6] = {0};

		ret = if_gethwaddr(&ifname[i*16], macaddr);
		if (ret || hwaddr_is_zero(macaddr)) {
			err("%s: %s hwaddr = 0!\n", __func__, &(ifname[i*16]));
			continue;
		}
		memcpy(p->albridge_macaddr[i*6], macaddr, 6);
		p->num_albridge++;
		dbg("IEEE1905 iface  = [%s] num_albridge = %d mac = ["MACFMT"]\n", &(ifname[i*16]), p->num_albridge, MAC2STR(p->albridge_macaddr[i*6]));

	}
	return ret;
}


//int topology_update_devinfo(struct topologyd_private *priv, struct node *n, struct tlv_device_info *tv0)
int topology_update_devinfo(struct topologyd_private *priv, struct node *n, struct tlv *t)
{
	struct tlv_device_info *devinfo =
				(struct tlv_device_info *)t->data;
	int j;
	uint8_t *ptr = t->data;
	int offset = sizeof(struct tlv_device_info);
	struct host_node *node;
	uint8_t node_intflist[MAX_IFNAME][6];
	int intf_num = 0;


	//dbg("%s %d we found a node\n", __func__, __LINE__);
	dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->hwaddr));

	node = host_node_lookup(priv->host.node_htable, n->hwaddr);
	if (!node)
		return -1;

	if ((node->ipv4addr_count >= HOST_MAX_IPV4ADDR) || (node->ipv6addr_count >= HOST_MAX_IPV6ADDR))
		return -1;

	n->local_intf_nbr = devinfo->num_interface;
	dbg("%s %d interface nr = %d\n", __func__, __LINE__, devinfo->num_interface);
	for (j = 0; j < devinfo->num_interface; j++) {
		struct local_interface *tif;

		ptr += offset;
		tif = (struct local_interface *)ptr;

		struct host_node *iface_node;

		dbg("%s %d interface nr = %d\n", __func__, __LINE__, devinfo->num_interface);
		memcpy(n->node_intf[j].interface_id, &tif->macaddr, 6);
		dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->node_intf[j].interface_id));


		/* Here we check so that the host can fill the media type for
		 * ieee1905 node also the interface node can be hidden
		 */
		iface_node = host_node_lookup(priv->host.node_htable, tif->macaddr);
		if (iface_node) {
			dbg("%s %d we found a node\n", __func__, __LINE__);

			if (((iface_node->ipv4addr_count) >= HOST_MAX_IPV4ADDR) &&
                                ((iface_node->ipv6addr_count) >= HOST_MAX_IPV6ADDR)) {
					err("ipv4 or ipv6 count is greater than max \n");
					return -1;
			}

			if (memcmp(node->hwaddr, iface_node->hwaddr, 6)) {
				int l = 0, m = 0, found1 = 0;
				int k = 0;

				if (!iface_node->is_device_mac)
					iface_node->ignore = true;
				memcpy(iface_node->al_node, node->hwaddr, HW_ADDR_LENGTH);
				//Here copy the ipv4 and ipv6 address
				//of the interface to the almac
				m = node->ipv4addr_count;
				for (l = 0; l < iface_node->ipv4addr_count; l++) {
					found1 = 0;
					for (k = 0; k < m; k++) {
						int ret1 = 0;
						ret1 = strncmp(node->ipv4addr_list[k], "", 24);
						if (ret1 == 0) {
							m--;
							break;
						}
						ret1 = strncmp(node->ipv4addr_list[k], iface_node->ipv4addr_list[l], 24);
						if (ret1 == 0)
							found1 = 1;
					}
					if (found1 == 0) {
						if (m >= HOST_MAX_IPV4ADDR)
							break;
						strncpy(node->ipv4addr_list[m], iface_node->ipv4addr_list[l], 24);
						m++;
					}
				}
				node->ipv4addr_count = m;
				m = node->ipv6addr_count;
				for (l = 0; l < iface_node->ipv6addr_count; l++) {
					found1 = 0;
					for (k = 0; k < m; k++) {
						int ret1 = 0;
						ret1 = strncmp(node->ipv6addr[k], "", 128);
						if (ret1 == 0) {
							m--;
							break;
						}
						ret1 = strncmp(node->ipv6addr[k], iface_node->ipv6addr[l], 128);
						if (ret1 == 0)
							found1 = 1;
					}
					if (found1 == 0) {
						if (m >= HOST_MAX_IPV6ADDR)
							break;
						strncpy(node->ipv6addr[m], iface_node->ipv6addr[l], 128);
						m++;
					}
				}
				node->ipv6addr_count = m;
			}
		}

		//Here we add the interfaces in the interfacelist
		memcpy(node_intflist[intf_num++], tif->macaddr, 6);


		n->node_intf[j].media_type = buf_get_be16((uint8_t *)&tif->mediatype);
		dbg("|%s:%d| Entry media_type %d\n", __func__, __LINE__, n->node_intf[j].media_type);
		//Here we by default make the interface power as on
		n->node_intf[j].power_status = 0;
		dbg("|%s:%d| Entry\n", __func__, __LINE__);
		//Here we also need to fill the media specific
		//data for the interface
		n->node_intf[j].media_specific_data_size = tif->sizeof_mediainfo;
		dbg("|%s:%d| Entry media size info %d \n", __func__, __LINE__, n->node_intf[j].media_specific_data_size);
		if (tif->sizeof_mediainfo == 10) {
			struct ieee80211_info *media80211 = (struct ieee80211_info *)tif->mediainfo;
			memcpy(n->node_intf[j].network_membership, &media80211->bssid, 6);
			n->node_intf[j].role = media80211->role;
			n->node_intf[j].ap_channel_band = media80211->ap_bandwidth;
			n->node_intf[j].freq_index1 = media80211->ap_channel_seg0_idx;
			n->node_intf[j].freq_index2 = media80211->ap_channel_seg1_idx;
		}

		offset = sizeof(struct local_interface) + tif->sizeof_mediainfo;
	}

	host_update_bridgenode(priv, n, intf_num, (uint8_t *)node_intflist);


	if (node->active == 0)
		host_change_active_state(priv, node, 1);

	return 0;
}

int topology_update_host_mediatype(struct node *n, struct topologyd_private *priv)
{
	struct host_node *node;
	struct host_node *iface_node;
	int res = 0;
	int i = 0;
	uint16_t media_type = 0;

	dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->hwaddr));
	node = host_node_lookup(priv->host.node_htable, n->hwaddr);
	if (!node)
		return -1;
	for (i = 0; i < n->local_intf_nbr; i++) {
		res = memcmp(n->nbr_localintf, n->node_intf[i].interface_id, 6);
		if (res != 0)
			continue;
		media_type = n->node_intf[i].media_type;
		if (media_type == I1905_802_3U_FAST_ETHERNET || media_type == I1905_802_3AB_GIGABIT_ETHERNET) {
			node->intf_type = HOST_TYPE_ETHER;
			node->is_copy = 1;
			node->is_copy_lan = 1;

		} else if (media_type == I1905_802_11B_2_4_GHZ || media_type == I1905_802_11G_2_4_GHZ ||
			media_type == I1905_802_11A_5_GHZ || media_type == I1905_802_11N_2_4_GHZ ||
			media_type == I1905_802_11N_5_GHZ || media_type == I1905_802_11AC_5_GHZ ||
			media_type == I1905_802_11AX_2_4_GHZ || media_type == I1905_802_11AX_5_GHZ ||
			media_type == I1905_802_11AD_60_GHZ || media_type == I1905_802_11AF_GHZ) {

			node->intf_type = HOST_TYPE_WIFI;
			node->is_copy = 1;
			node->is_copy_lan = 0;

		} else {

			err("|%s:%d|unknown media type \n ", __func__, __LINE__);
			return -1;
		}
		iface_node = host_node_lookup(priv->host.node_htable, n->nbr_localintf);
		if (iface_node) {
			if (node->intf_type == HOST_TYPE_WIFI)
				memcpy(&node->stats, &iface_node->stats, sizeof(node->stats));
		}
		//set the interface in config file
		host_set_interface_type(node, priv);
		break;
	}

	return 0;
}

int topology_update_neigh_list(struct topologyd_private *priv, struct node *n,
		struct tlv_1905neighbor *tv0, uint16_t tlv_len, int *list_num)
{
	int j = 0;
	int size = 0;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);
	dbg("|%s:%d| Inside tlv len is %d \n", __func__, __LINE__, tlv_len);
	memcpy(n->nbr_localintf, &tv0->local_macaddr, 6);
	dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->nbr_localintf));
	size = tlv_len - (sizeof(struct tlv_1905neighbor));
	size = size / (sizeof(struct i1905_neighbor));
	dbg("|%s:%d| Inside 1905nbr %d\n", __func__, __LINE__, size);
	for (j = 0; j < size; j++) {
		int ret = 0;

		memcpy(n->nbr_list[*list_num].nbr_macid, &tv0->nbr[j].aladdr, 6);
		dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->nbr_list[*list_num].nbr_macid));
		ret = memcmp(priv->ieee1905_macaddr, &tv0->nbr[j].aladdr, 6);
		if (ret == 0)
			topology_update_host_mediatype(n, priv);
		*list_num = *list_num + 1;
		n->nbr_num = *list_num;
		dbg("|%s:%d| Inside 1905nbr %d\n", __func__, __LINE__, n->nbr_num);
	}
	return 0;
}

int topology_update_non_i1905nbr_list(struct node *n, struct tlv_non1905_neighbor *tv0,
		uint16_t tlv_len, int *list_num)
{
	int j = 0;
	int size = 0;
	dbg("|%s:%d| Inside\n", __func__, __LINE__);
	dbg("|%s:%d| Inside tlv len is %d \n", __func__, __LINE__, tlv_len);
	memcpy(n->non1905_nbr_localintf, &tv0->local_macaddr, 6);
	dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->non1905_nbr_localintf));
	size = tlv_len - (sizeof(struct tlv_non1905_neighbor));
	size = size / (sizeof(struct non1905_neighbor));
	dbg("the size is %d", size);
	//dbg("|%s:%d| Inside non1905nbr %d\n", __func__, __LINE__, n->non1905_nbr_num);
	for (j = 0; j < size; j++) {
		memcpy(n->non1905_nbrlist[*list_num], &tv0->non1905_nbr[j].macaddr, 6);
		dbg("|%s:%d| Entry"MACFMT"\n", __func__, __LINE__, MAC2STR(n->non1905_nbrlist[*list_num]));
		*list_num = *list_num + 1;
		n->non1905_nbr_num = *list_num;
		dbg("|%s:%d| Inside non1905nbr %d\n", __func__, __LINE__, n->non1905_nbr_num);
	}
	return 0;
}

int topology_update_l2_neigh(struct node *n)
{
	//TODO not in i1905 stack
	return 0;
}

int topology_update_bridging_info(struct node *n, struct tlv_device_bridge_caps *tv0)
{
	int j = 0, k = 0;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);
	n->bridge_tuple_num = tv0->num_tuples;
	dbg(" bridging tuple number [%d]\n", tv0->num_tuples);

	for (j = 0; j < tv0->num_tuples; j++) {
		n->br_tuple[j].tuple_num = tv0->tuple[j].num_macaddrs;

		for (k = 0; k < tv0->tuple[j].num_macaddrs; k++) {
			dbg(" bridge_num[%d] br_mac_num[%d]\n", tv0->num_tuples, tv0->tuple[j].num_macaddrs);
			memcpy(n->br_tuple[j].tuple_macid[k], &tv0->tuple[j].addr[k].macaddr, 6);
			dbg("|%s:%d| Entry bridge mac is "MACFMT"\n", __func__, __LINE__, MAC2STR(n->br_tuple[j].tuple_macid[k]));
		}
	}
	return 0;
}

int topology_update_poweroff_list(struct node *n)
{
	//TODO not in i1905 stack
	return 0;
}

void topologyd_process_topology_response(struct cmdu_buff *cstruct, struct topologyd_private *priv, char *itfr_name)
{

	struct node n = { 0 };
	int ret;
	struct tlv *tv[11][TLV_MAXNUM] = {0};

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("|%s:%d| Inside \n", __func__, __LINE__);

	if (!validate_topology_response(cstruct, tv)) {
		dbg("cmdu validation: [TOPOLOGY_RESPONSE] failed\n");
		return;
	}

	if (!tv[0][0]) {
		err("mandatory tlv not present\n");
		return;
	}
	if (tv[0][0]) {
		struct tlv_device_info *devinfo =
			(struct tlv_device_info *)tv[0][0]->data;
		dbg("|%s:%d| Inside ALMAC is"MACFMT"\n", __func__, __LINE__, MAC2STR(devinfo->aladdr));
		if (hwaddr_is_zero(devinfo->aladdr)) {
			fprintf(stderr, "%s: Discard topo response from aladdr = 0!\n",
				__func__);

			return;
		}

		memcpy(&n.hwaddr, devinfo->aladdr, 6);
		if (itfr_name != NULL)
			strncpy(n.ingress_ifr_name, itfr_name, strlen(itfr_name));
		ret = topology_update_devinfo(priv, &n, tv[0][0]);
	}

	if (tv[1][0]) {
		int num = 0;

		while (num < TLV_MAXNUM && tv[1][num]) {
			ret = topology_update_bridging_info(&n,
					(struct tlv_device_bridge_caps *)tv[1][num]->data);
			if (ret)
				break;
			num++;
		}
	}


	if (tv[2][0]) {
		int num = 0;
		int list_num = 0;

		while (num < TLV_MAXNUM && tv[2][num]) {
			ret = topology_update_non_i1905nbr_list(&n,
						(struct tlv_non1905_neighbor *)tv[2][num]->data,
						BUF_GET_BE16(tv[2][num]->len),
						&list_num);
			if (ret)
				break;
			num++;
		}
	}

	if (tv[3][0]) {
		int num = 0;
		int list_num = 0;
		while (num < TLV_MAXNUM && tv[3][num]) {
			ret = topology_update_neigh_list(priv, &n,
						(struct tlv_1905neighbor *)tv[3][num]->data,
						BUF_GET_BE16(tv[3][num]->len),
						&list_num);
			if (ret)
				break;
			num++;
		}
	}
	if (tv[4][0]) {
		int num = 0;
		while (num < TLV_MAXNUM && tv[4][num]) {
			ret = topology_update_poweroff_list(&n);
			if (ret)
				break;
			num++;
		}
	}

	if (tv[5][0]) {
		int num = 0;
		while (num < TLV_MAXNUM && tv[5][num]) {
			ret = topology_update_l2_neigh(&n);
			if (ret)
				break;
			num++;
		}
	}

	if (tv[8][0]) {
		ret = topology_update_host_assoc_client(priv,
					(struct tlv_assoc_client *)tv[8][0]->data);
	}

	if (tv[7][0]) {
		/*If AP Operational BSS TLV the device supports easy mesh*/
		n.is_easymesh = 1;

	}

	dbg("Updating Node" MACFMT " from topology response\n",
		MAC2STR(n.hwaddr));
	topologyd_node_add(priv, &n);
}

int topology_update_device_ident(struct node *n, struct tlv_device_identification *ident)
{
	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	memset(n->friendly_name, 0, 64);
	memcpy(n->friendly_name, ident->name, 64);
	n->friendly_name[63] = '\0';

	memset(n->manufacturer_name, 0, 64);
	memcpy(n->manufacturer_name, ident->manufacturer, 64);
	n->manufacturer_name[63] = '\0';

	memset(n->manufacturer_model, 0, 64);
	memcpy(n->manufacturer_model, ident->model, 64);
	n->manufacturer_model[63] = '\0';

	return 0;
}

int topology_update_control_url(struct node *n, struct tlv_control_url *ctrl, int len)
{
	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	if (len >= 1024)
		return -1;

	memcpy(n->control_url, ctrl->url, len);

	return 0;
}

int topology_update_ipv4(struct node *n, struct tlv *t, uint8_t *hwaddr,
		struct topologyd_private *priv)
{
	struct tlv_ipv4 *ip = (struct tlv_ipv4 *)t->data;
	int offset = sizeof(struct tlv_ipv4);
	uint8_t *ptr = t->data;
	int i, j;
	int cnt = 0;

	for (i = 0; i < ip->num_interfaces; i++) {
		struct ipv4_interface *tif;

		ptr += offset;
		tif = (struct ipv4_interface *)ptr;
		offset = sizeof(struct ipv4_interface);

		for (j = 0; j < tif->num_ipv4; j++) {
			struct ipv4_entry *e;
			struct host_node *iface_node;

			ptr += offset;
			e = (struct ipv4_entry *)ptr;

			memcpy(n->device_ipv4[cnt].mac_address, tif->macaddr, 6);
			iface_node = host_node_lookup(priv->host.node_htable, tif->macaddr);

			if (!iface_node) {
				iface_node = host_topo_node_add(priv, 0, tif->macaddr, 0);
				if (!iface_node)
					err("Failed to add node in the hosts lookup\n");
			}

			if (iface_node) {
				struct host_node *node;
				int rv = -1;
				char mac_str[18] = { 0 };

				dbg("%s %d we found a node in ipv4\n", __func__, __LINE__);
				node = host_node_lookup(priv->host.node_htable, hwaddr);
				if (node) {
					rv = memcmp(node->hwaddr, iface_node->hwaddr, 6);
					if (rv != 0) {
						node->ignore = true;
						iface_node->ignore = false;
						memcpy(iface_node->al_node, node->hwaddr, HW_ADDR_LENGTH);
						iface_node->intf_type = node->intf_type;
						hwaddr_ntoa(iface_node->hwaddr, mac_str);
						config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type",
							(iface_node->intf_type == HOST_TYPE_WIFI ? "wifi" : "eth"),
							priv->host_global_cfg.persistent);
						iface_node->is_device_mac = 1;
						//Here we need a variable so that the node value is not changed
						//periodically
						node->is_copy = 1;
						//Here we need to copy the data of the br-lan
						//to that of the al-mac node
						host_copy_node_data(iface_node, node, priv);
						if (node->is_event == 0) {
							host_send_client_event(priv, node, 1);
							node->is_event = 1;
						}
					} else {
						node->is_copy = 0;
					}
				}
			}
			memcpy(n->device_ipv4[cnt].ipv4_addr, e->address, sizeof(struct in_addr));
			n->device_ipv4[cnt].ipv4_addr_type = e->type;
			memcpy(n->device_ipv4[cnt].dhcp_addr, e->dhcpserver, sizeof(struct in_addr));

			offset = sizeof(struct ipv4_entry);
			cnt++;
		}
	}
	n->ipv4_intf_nbr = cnt;
	return 0;
}


int topology_update_ipv6(struct node *n, struct tlv *t)
{
	struct tlv_ipv6 *ip = (struct tlv_ipv6 *)t->data;
	int offset = sizeof(struct tlv_ipv6);
	uint8_t *ptr = t->data;
	int i, j;
	int cnt = 0;

	for (i = 0; i < ip->num_interfaces; i++) {
		struct ipv6_interface *tif;

		ptr += offset;
		tif = (struct ipv6_interface *)ptr;
		offset = sizeof(struct ipv6_interface);

		for (j = 0; j < tif->num_ipv6; j++) {
			struct ipv6_entry *e;

			ptr += offset;
			e = (struct ipv6_entry *)ptr;

			memcpy(n->device_ipv6[cnt].mac_address, tif->macaddr, 6);
			//TODO: linklocal address
			memcpy(n->device_ipv6[cnt].ipv6_addr, e->address, sizeof(struct in6_addr));
			n->device_ipv6[cnt].ipv6_addr_type = e->type;
			memcpy(n->device_ipv6[cnt].ipv6addr_origin, e->origin, sizeof(struct in6_addr));
			dbg("Inside topologyd ipv6 hhl message intf_num[%d] ipv6_num[%0x]\n",
				ip->num_interfaces, tif->num_ipv6);
			offset = sizeof(struct ipv6_entry);
			cnt++;
		}
	}
	n->ipv6_intf_nbr = cnt;
	return 0;
}

void topologyd_process_higherlayer_response(struct cmdu_buff *cstruct, struct topologyd_private *priv)
{
	struct node *n = NULL;
	int ret;
	struct tlv *tv[6][TLV_MAXNUM] = {0};
	uint8_t aladdr_origin[6] = {0};
	struct tlv_1905_profile *profile;

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	if (!validate_higher_layer_response(cstruct, tv)) {
		dbg("cmdu validation: [HIGHER_LAYER_RESPONSE] failed\n");
		return;
	}

	if (!tv[0][0] || !tv[1][0] || !tv[2][0])
		return;

	memcpy(aladdr_origin, tv[0][0]->data, t_tlv_length(tv[0][0]));
	if (hwaddr_is_zero(aladdr_origin)) {
		fprintf(stderr,
		"%s: Discard higherlayer response from aladdr = 0!\n",
		__func__);

		return;
	}

	//Here we need to check the aladdr is selfnode addr then return
	ret = memcmp(aladdr_origin, priv->selfnode.hwaddr, 6);
	if (ret == 0) {
		//dbg(" The HL resp corresponds to the self node we are ignoring this details\n");
		//return;
		n = &priv->selfnode;
	} else {
		n = node_lookup(priv->topo.node_htable, aladdr_origin);
		if (n == NULL) {
			err("Node not found \n");
			return;
		}
	}

	if (tv[1][0]) {
		profile = (struct tlv_1905_profile *)tv[1][0]->data;
		if (profile->version != PROFILE_1905_1 &&
			profile->version != PROFILE_1905_1A) {

				fprintf(stderr,
					"%s: Discard higherlayer response (invalid profile)!\n",
					__func__);
				return;
		}
		n->version = profile->version;
	}

	if (tv[2][0]) {
		ret = topology_update_device_ident(n, (struct tlv_device_identification *)tv[2][0]->data);
	}

	if (tv[3][0]) {
		ret = topology_update_control_url(n, (struct tlv_control_url *)tv[3][0]->data, BUF_GET_BE16(tv[3][0]->len));
	}

	if (tv[4][0]) {
		ret = topology_update_ipv4(n, tv[4][0], aladdr_origin, priv);
	}

	if (tv[5][0]) {
		ret = topology_update_ipv6(n, tv[5][0]);
	}
}

void topologyd_process_autoconfiguration_response(struct cmdu_buff *cstruct, struct topologyd_private *priv)
{

	struct tlv_supported_band *freq;
	struct tlv *tv[4][TLV_MAXNUM] = {0};

	dbg("%s -------------->\n", __func__);

	if (!validate_ap_autoconfig_response(cstruct, tv)) {
		dbg("cmdu validation: [AP_AUTOCONFIG_RESPONSE] failed\n");
		return;
	}

	if (!tv[0][0] || !tv[3][0])
		return;
	if (tv[0][0]->data[0] != IEEE80211_ROLE_REGISTRAR) {
		dbg("%s: Discard ap-autoconfig response for role != registrar\n",
			__func__);
		return;
	}

	freq = (struct tlv_supported_band *)tv[3][0]->data;
	if (freq->band > IEEE80211_FREQUENCY_BAND_60_GHZ) {
		dbg("%s: Discard ap-autoconfig response for invalid WiFi band\n",
			 __func__);
		return;
	}

	if (freq->band >= IEEE80211_FREQUENCY_BAND_2_4_GHZ &&
		freq->band <= IEEE80211_FREQUENCY_BAND_60_GHZ) {

		g_registrar_freq_band[freq->band] = 1;
	}
}

static int topologyd_refresh(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	if (!priv->algo_running)
		topologyd_algo_run(priv);

	return 0;
}

int topology_update_tx_link_metric(struct node *n, struct tlv_tx_linkmetric *txl, int len)
{
	uint8_t fr_al_mac_id[6];
	uint8_t to_al_mac_id[6];
	int i = 0, j = 0;
	int ret = 0;
	int size = 0;
	struct tx_link_info *txlinfo;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	memcpy(fr_al_mac_id, txl->aladdr, 6);
	memcpy(to_al_mac_id, txl->neighbor_aladdr, 6);

	size = len - (sizeof(struct tlv_tx_linkmetric));
	size = size / (sizeof(struct tx_link_info));
	dbg("the size of nbr list is %d \n", size);
	for (i = 0; i < n->nbr_num; i++) {
		ret = memcmp(n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
		if (ret != 0)
			continue;
		n->nbr_list[i].metric_num = size;
		dbg(" Inside the transmitter link metric tlv i = [%d] ln_metric_num = [%d]\n", i, n->nbr_list[i].metric_num);
		for (j = 0; j < size; j++) {
			txlinfo = (struct tx_link_info *)&txl->link[j];
			memcpy(n->nbr_list[i].ln_metric[j].nbr_ifr_addr, txlinfo->neighbor_macaddr, 6);
			dbg(" Inside the NEW transmitter link metric tlv nbr mac is["MACFMT"]\n", MAC2STR(txlinfo->neighbor_macaddr));
			n->nbr_list[i].ln_metric[j].ieee802dot1br = txlinfo->has_bridge;
			n->nbr_list[i].ln_metric[j].tx_packet_errors = BUF_GET_BE32(txlinfo->errors);
			n->nbr_list[i].ln_metric[j].tx_packet_ok = BUF_GET_BE32(txlinfo->packets);
			n->nbr_list[i].ln_metric[j].tx_max_xput = BUF_GET_BE16(txlinfo->max_throughput);
			n->nbr_list[i].ln_metric[j].tx_link_availability = BUF_GET_BE16(txlinfo->availability);
			n->nbr_list[i].ln_metric[j].tx_phy_rate = BUF_GET_BE16(txlinfo->phyrate);
		}
	}
	return 0;
}

int topology_update_rx_link_metric(struct node *n, struct tlv_rx_linkmetric *rxl, int len)
{
	uint8_t fr_al_mac_id[6];
	uint8_t to_al_mac_id[6];
	int i = 0, j = 0;
	int ret = 0;
	int size = 0;
	struct rx_link_info *rxlinfo;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	memcpy(fr_al_mac_id, rxl->aladdr, 6);
	memcpy(to_al_mac_id, rxl->neighbor_aladdr, 6);

	size = len - (sizeof(struct tlv_rx_linkmetric));
	size = size / (sizeof(struct rx_link_info));
	dbg("the size of nbr list is %d \n", size);
	for (i = 0; i < n->nbr_num; i++) {
		ret = memcmp(n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
		if (ret != 0)
			continue;
		n->nbr_list[i].metric_num = size;
		dbg(" Inside the receiver link metric tlv  ln_metric_num %d\n", size);
		for (j = 0; j < size; j++) {
			rxlinfo = (struct rx_link_info *)&rxl->link[j];
			n->nbr_list[i].ln_metric[j].rx_packet_errors = BUF_GET_BE32(rxlinfo->errors);
			n->nbr_list[i].ln_metric[j].rx_packet_ok = BUF_GET_BE32(rxlinfo->packets);
			n->nbr_list[i].ln_metric[j].rx_rssi = rxlinfo->rssi;
		}
	}
	return 0;
}

void topologyd_process_linkmetric_response(struct cmdu_buff *cstruct, struct topologyd_private *priv)
{

	struct tlv *tv[2][TLV_MAXNUM] = {0};
	int num = 0;
	int ret;
	struct node *n;

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	if (!validate_link_metric_response(cstruct, tv)) {
		dbg("cmdu validation: [LINK_LAYER_RESPONSE] failed\n");
		return;
	}

	while (num < TLV_MAXNUM && tv[0][num]) {
		struct tlv_tx_linkmetric *txl =
			(struct tlv_tx_linkmetric *)tv[0][num]->data;

		if (hwaddr_is_zero(txl->aladdr) ||
			hwaddr_is_zero(txl->neighbor_aladdr)) {

			fprintf(stderr, "%s: Discard Tx-link response\n",
				__func__);
			return;
		}

		ret = memcmp(txl->aladdr, priv->selfnode.hwaddr, 6);
		if (ret == 0) {
			// dbg(" The Link Layer resp is the self node we are ignoring this details\n");
			// return;
			n = &priv->selfnode;
		} else {

			n = node_lookup(priv->topo.node_htable, txl->aladdr);

			//If we dont find the node in our devices
			if (n == NULL) {
				err("Node not found \n");
				return;
			}
		}

		ret = topology_update_tx_link_metric(n,
				(struct tlv_tx_linkmetric *)tv[0][num]->data,
				BUF_GET_BE16(tv[0][num]->len));
		if (ret)
			break;
		num++;
	}

	num = 0;
	while (num < TLV_MAXNUM && tv[1][num]) {
		struct tlv_rx_linkmetric *rxl =
			(struct tlv_rx_linkmetric *)tv[1][num]->data;

		if (hwaddr_is_zero(rxl->aladdr) ||
			hwaddr_is_zero(rxl->neighbor_aladdr)) {

			fprintf(stderr, "%s: Discard Tx-link response\n",
				__func__);

			return;
		}

		ret = memcmp(rxl->aladdr, priv->selfnode.hwaddr, 6);
		if (ret == 0) {
			//dbg(" The Link Layer resp is the self node we are ignoring this details\n");
			//return;
			n = &priv->selfnode;
		} else {

			n = node_lookup(priv->topo.node_htable, rxl->aladdr);
			//If we dont find the node in our devices
			if (n == NULL) {
				err("Node not found \n");
				return;
			}
		}

		ret = topology_update_rx_link_metric(n,
				(struct tlv_rx_linkmetric *)tv[1][num]->data,
				BUF_GET_BE16(tv[1][num]->len));
		if (ret)
			break;
		num++;
	}
	return;
}


void topologyd_process_topology_discovery(struct cmdu_buff *cstruct,
						struct topologyd_private *priv)
{
	struct host_node *node;
	struct tlv *tv[2][TLV_MAXNUM] = {0};
	struct tlv_macaddr *macaddr;
	struct tlv_aladdr *aladdr;

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	if (!validate_topology_discovery(cstruct, tv)) {
		dbg("cmdu validation: [TOPOLOGY_DISCOVERY] failed\n");
		return;
	}

	if (!tv[0][0] || !tv[1][0])
		return;

	aladdr = (struct tlv_aladdr *) tv[0][0]->data;
	macaddr = (struct tlv_macaddr *) tv[1][0]->data;

	node = host_node_lookup(priv->host.node_htable, aladdr->macaddr);
	if (!node) {
		dbg("|%s:%d| did not find node " MACFMT "\n", __func__,
		    __LINE__, MAC2STR(aladdr->macaddr));
		return;
	}

	memcpy(node->link_hwaddr, macaddr->macaddr, 6);
	dbg("|%s:%d| added link addr " MACFMT " for node " MACFMT "\n", __func__,
	    __LINE__, MAC2STR(node->link_hwaddr), MAC2STR(node->hwaddr));
}


void topologyd_event_handler(void *c, struct blob_attr *msg)
{
	struct topologyd_private *priv = (struct topologyd_private *)c;
	int ret = 0;
	struct cmdu_buff *cstruct = NULL;
	uint16_t msg_type = 0;
	uint16_t msg_mid = 0;
	uint8_t *tlv = NULL;
	uint32_t len = 0;
	char itfr_name[16] = { 0 };
	uint8_t source[6] = { 0 }, origin[6] = { 0 };


	dbg("Inside %s...\n", __func__);
	if (!msg)
		return;

	//Calling the function to get the args from json msg

	ret = ieee1905_msg_from_ubus(msg, source, origin, &msg_type, &msg_mid,
				     &tlv, &len, itfr_name);
	if (ret != 0) {
		err("ieee1905_msg_from_ubus error\n");
		if (tlv != NULL)
			free(tlv);
		return;
	}

	dbg("Inside topology event handler %hu %hu  %s", msg_type, msg_mid, itfr_name);
	if (msg_type != CMDU_TYPE_TOPOLOGY_RESPONSE &&
	    msg_type != CMDU_TYPE_TOPOLOGY_NOTIFICATION &&
	    msg_type != CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE &&
	    msg_type != CMDU_TYPE_HIGHER_LAYER_RESPONSE &&
	    msg_type != CMDU_TYPE_LINK_METRIC_RESPONSE &&
	    msg_type != CMDU_TYPE_GENERIC_PHY_RESPONSE &&
	    msg_type != CMDU_TYPE_TOPOLOGY_QUERY &&
	    msg_type != CMDU_TYPE_HIGHER_LAYER_QUERY &&
	    msg_type != CMDU_TYPE_LINK_METRIC_QUERY &&
	    msg_type != CMDU_TYPE_TOPOLOGY_DISCOVERY) {
		if (tlv != NULL)
			free(tlv);
		return;
	}

	if (msg_type != CMDU_TYPE_TOPOLOGY_QUERY &&
	    msg_type != CMDU_TYPE_HIGHER_LAYER_QUERY &&
	    msg_type != CMDU_TYPE_LINK_METRIC_QUERY &&
	    msg_type != CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE &&
	    msg_type != CMDU_TYPE_TOPOLOGY_NOTIFICATION &&
	    msg_type != CMDU_TYPE_TOPOLOGY_DISCOVERY) {

		dbg("Inside %s...DEQUEUE msg_type = [%d] msg_mid = [%d] src =["MACFMT"]\n", __func__,
			msg_type, msg_mid, MAC2STR(origin));
		ret = t_cmdu_ackq_dequeue(&priv->cmdu_ack_q, msg_type,
					msg_mid, origin, NULL);
		if (ret) {
			err("topologyd: drop unexpected CMDU (mid = %d)\n", msg_mid);
			if (tlv != NULL)
				free(tlv);
			return;
		}
	}

	cstruct = t_cmdu_alloc_custom(msg_type, &msg_mid, NULL, source, tlv, len);
	if (cstruct == NULL) {
		if (tlv != NULL)
			free(tlv);
		return;
	}

	dbg("%s: cmdu_alloc_custom() succeeded! cmdu->cdata->hdr.mid %u\n",
	    __func__, t_cmdu_get_mid(cstruct));

	dbg("After Event handler\n");

	switch (msg_type) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
	{
		if (*tlv == MAP_TLV_MULTIAP_PROFILE) {
			info("ignore the query with multi-ap tlv\n");
			goto out;
		}
		/* NO-OP */
	}
	break;
	case CMDU_TYPE_HIGHER_LAYER_QUERY:
	case CMDU_TYPE_LINK_METRIC_QUERY:
	{
		dbg("Received  query  message\n");
		topology_send_query_response(priv, msg_type, origin, source, msg_mid, itfr_name);
	}
	break;
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
	{
		dbg("Received topology response  message\n");
		topologyd_process_topology_response(cstruct, priv, itfr_name);
	}
	break;
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
	{
		/*Here we recieved the topology notification */
		/* we muct run the algo to update the topology*/
		if (!priv->algo_running)
			topologyd_algo_run(priv);
		topologyd_process_topology_notification(cstruct, priv);
	}
	break;
	case CMDU_TYPE_HIGHER_LAYER_RESPONSE:
	{
		dbg("Received higher layer response message\n");
		topologyd_process_higherlayer_response(cstruct, priv);
	}
	break;
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
	{
		dbg("Received ap autoconfiguration response message\n");
		topologyd_process_autoconfiguration_response(cstruct, priv);
	}
	break;
	case CMDU_TYPE_LINK_METRIC_RESPONSE:
	{
		dbg("Received link metric response\n");
		topologyd_process_linkmetric_response(cstruct, priv);

	}
	break;
	case CMDU_TYPE_GENERIC_PHY_RESPONSE:
	{
		dbg("Received a generic phy response\n");
		//TODO not present in i1905
		//topologyd_process_generic_phy_response(cstruct, priv);
	}
	break;
	case CMDU_TYPE_TOPOLOGY_DISCOVERY:
	{
		dbg("Received topology response  message\n");
		topologyd_process_topology_discovery(cstruct, priv);
	}
	break;
	default:
		dbg("Event handler received default message\n");
	break;
	}
out:
	if (tlv != NULL)
		free(tlv);
	if (cstruct != NULL)
		t_cmdu_free(cstruct);
}

#if 0
void topology_add_txrx_data(uint8_t *metric, struct topologyd_private *priv)
{
	uint8_t fr_al_mac_id[6];
	uint8_t to_al_mac_id[6];
	uint32_t i, j, res;
	struct node *n = NULL;

	if (metric == NULL)
		return;

	dbg("Updating Node with the Link layer response message\n");
	//Here we 1st compare that the node is self node

	if (*metric == TLV_TYPE_TRANSMITTER_LINK_METRIC) {
		dbg(" Inside the transmitter link metric tlv\n");
		struct transmitterLinkMetricTLV *ptr;

		ptr = (struct transmitterLinkMetricTLV *)metric;

		memcpy(fr_al_mac_id, ptr->local_al_address, 6);
		memcpy(to_al_mac_id, ptr->neighbor_al_address, 6);

		res = memcmp(fr_al_mac_id, priv->selfnode.hwaddr, 6);
		if (res == 0) {
			dbg(" The Link Layer resp corresponds to the self node we are ignoring this details\n");
			return;
		}

		n = node_lookup(priv->topo.node_htable, fr_al_mac_id);

		//If we find the node in our devices
		if (n != NULL) {
			//getting the node specific params then we need to search the neighbor for which this metric is given
			for (i = 0; i < n->nbr_num; i++) {
				res = memcmp(n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
				if (res != 0)
					continue;
				n->nbr_list[i].metric_num = ptr->transmitter_link_metrics_nr;
				dbg(" Inside the transmitter link metric tlv  ln_metric_num %d\n", ptr->transmitter_link_metrics_nr);
				for (j = 0; j < ptr->transmitter_link_metrics_nr; j++) {
					memcpy(n->nbr_list[i].ln_metric[j].nbr_ifr_addr, ptr->transmitter_link_metrics[j].neighbor_interface_address, 6);
					n->nbr_list[i].ln_metric[j].ieee802dot1br = ptr->transmitter_link_metrics[j].bridge_flag;
					n->nbr_list[i].ln_metric[j].tx_packet_errors = ptr->transmitter_link_metrics[j].packet_errors;
					n->nbr_list[i].ln_metric[j].tx_packet_ok = ptr->transmitter_link_metrics[j].transmitted_packets;
					n->nbr_list[i].ln_metric[j].tx_max_xput = ptr->transmitter_link_metrics[j].mac_throughput_capacity;
					n->nbr_list[i].ln_metric[j].tx_link_availability = ptr->transmitter_link_metrics[j].link_availability;
					n->nbr_list[i].ln_metric[j].tx_phy_rate = ptr->transmitter_link_metrics[j].phy_rate;
				}
			}
		}
	} else if (*metric == TLV_TYPE_RECEIVER_LINK_METRIC) {

		struct receiverLinkMetricTLV *p;

		dbg(" Inside the receiver link metric tlv\n");

		p = (struct receiverLinkMetricTLV *)metric;

		memcpy(fr_al_mac_id, p->local_al_address, 6);
		memcpy(to_al_mac_id, p->neighbor_al_address, 6);

		res = memcmp(fr_al_mac_id, priv->selfnode.hwaddr, 6);
		if (res == 0) {
			dbg(" The Link Layer resp corresponds to the self node we are ignoring this details\n");
			return;
		}

		n = node_lookup(priv->topo.node_htable, fr_al_mac_id);
		//If we find the node in our devices
		if (n != NULL) {
			//getting the node specific params then we need to search the neighbor for which this metric is given
			for (i = 0; i < n->nbr_num; i++) {
				res = memcmp(n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
				if (res != 0)
					continue;
				dbg(" Inside the receiver link metric  ln_num is %d\n", p->receiver_link_metrics_nr);
				for (j = 0; j < p->receiver_link_metrics_nr; j++) {
					n->nbr_list[i].ln_metric[j].rx_packet_errors = p->receiver_link_metrics[j].packet_errors;
					n->nbr_list[i].ln_metric[j].rx_packet_ok = p->receiver_link_metrics[j].packets_received;
					n->nbr_list[i].ln_metric[j].rx_rssi = p->receiver_link_metrics[j].rssi;
				}
			}
		}
	}
}

void topologyd_process_generic_phy_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv)
{
	int i = 0;
	uint8_t *p;

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("Inside Geniric Phy response\n");
	while (NULL != (p = cstruct->tlv_cstruct_list[i])) {

		switch (*p) {
		case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			dbg(" Inside generic_phy_device_info\n");
			topology_add_generic_phy_data(p, priv);
			//struct genericPhyDeviceInformationTypeTLV t = (struct genericPhyDeviceInformationTypeTLV *)p;

		}
		break;
		case TLV_TYPE_VENDOR_SPECIFIC:
		{
			//Nothing to do with the vendor tlv now
		}
		break;
		}
		i++;
	}
}

void topology_add_generic_phy_data(uint8_t *phy, struct topologyd_private *priv)
{
	uint8_t fr_al_mac_id[6];
	uint32_t i, j, res;
	struct node *n = NULL;

	if (phy == NULL)
		return;

	err("Updating Node with the generic phy response message\n");

	if (*phy == TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION) {
		dbg(" Inside the generic phy info tlv\n");
		struct genericPhyDeviceInformationTypeTLV *ptr;

		ptr = (struct genericPhyDeviceInformationTypeTLV *)phy;

		memcpy(fr_al_mac_id, ptr->al_mac_address, 6);

		res = memcmp(fr_al_mac_id, priv->selfnode.hwaddr, 6);
		if (res == 0) {
			dbg(" The generic phy resp corresponds to the self node we are ignoring this details\n");
			return;
		}

		n = node_lookup(priv->topo.node_htable, fr_al_mac_id);

		//If we find the node in our devices
		if (n != NULL) {
			//getting the interface for which this information is given
			for (i = 0; i < ptr->local_interfaces_nr; i++) {
				for (j = 0; j < n->local_intf_nbr; j++) {

					res = memcmp(ptr->local_interfaces[i].local_interface_address, n->node_intf[j].interface_id, 6);
					if (res != 0)
						continue;
					memcpy(n->node_intf[j].generic_phy_oui, ptr->local_interfaces[i].generic_phy_common_data.oui, 3);
					n->node_intf[j].generic_phy_variant =  ptr->local_interfaces[i].generic_phy_common_data.variant_index;
					if (ptr->local_interfaces[i].generic_phy_description_xml_url_len > 0)
						memcpy(n->node_intf[j].generic_phy_url, ptr->local_interfaces[i].generic_phy_description_xml_url, ptr->local_interfaces[i].generic_phy_description_xml_url_len);
				}
			}
		}
	}
}
#endif
void topologyd_dump_node(struct blob_buf *bb, struct node *p, int is_self)
{
	char mac_str[18] = { 0x0 };
	char freq_str[50] = { 0x0 };
	char behind_mac_str[1024] = { 0x0 };
	void *nbr_array, *non1905_array = NULL, *interfaces = NULL, *table = NULL, *l2_neighbor = NULL;
	void *bridge_tuple = NULL, *ipv4_param = NULL, *ipv6_param = NULL, *link_array = NULL;
	void *bridge_tuple_mac = NULL;
	void *table1 = NULL;
	int  j, k, flag = 0, i = 0;
	char addr[16] = {0};
	char addr_v6[40] = {0};
	char oui[10] = {0};
	char phy_variant[10] = {0};

	hwaddr_ntoa(p->hwaddr, mac_str);
	blobmsg_add_string(bb, "ieee1905_macaddr", mac_str);
	blobmsg_add_u32(bb, "num_neighbors", p->nbr_num);

	nbr_array = blobmsg_open_array(bb, "neighbors");
	for (j = 0; j < p->nbr_num; j++) {
		table1 = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->nbr_list[j].nbr_macid, mac_str);
		blobmsg_add_string(bb, "nbr_mac_id", mac_str);
		if (is_self == 0) {
			link_array = blobmsg_open_array(bb, "link_metrics");
			dbg(" Inside the print link almac = [%s] ln_metric_num = [%d]\n", mac_str, p->nbr_list[j].metric_num);
			for (k = 0; k < p->nbr_list[j].metric_num; k++) {
				table = blobmsg_open_table(bb, NULL);
				hwaddr_ntoa(p->nbr_list[j].ln_metric[k].nbr_ifr_addr, mac_str);
				blobmsg_add_string(bb, "nbr_intf_macid", mac_str);
				blobmsg_add_u32(bb, "ieee_bridge", p->nbr_list[j].ln_metric[k].ieee802dot1br);
				blobmsg_add_u32(bb, "tx_packet_error", p->nbr_list[j].ln_metric[k].tx_packet_errors);
				blobmsg_add_u32(bb, "tx_packet_ok", p->nbr_list[j].ln_metric[k].tx_packet_ok);
				blobmsg_add_u32(bb, "mac_thrput_capacity", p->nbr_list[j].ln_metric[k].tx_max_xput);
				blobmsg_add_u32(bb, "link_availability", p->nbr_list[j].ln_metric[k].tx_link_availability);
				blobmsg_add_u32(bb, "phy_rate", p->nbr_list[j].ln_metric[k].tx_phy_rate);
				blobmsg_add_u32(bb, "rx_packet_error", p->nbr_list[j].ln_metric[k].rx_packet_errors);
				blobmsg_add_u32(bb, "rx_packet_ok", p->nbr_list[j].ln_metric[k].rx_packet_ok);
				blobmsg_add_u32(bb, "rssi", p->nbr_list[j].ln_metric[k].rx_rssi);
				blobmsg_close_table(bb, table);
			}
			blobmsg_close_array(bb, link_array);
		}
		blobmsg_close_table(bb, table1);
	}
	blobmsg_close_array(bb, nbr_array);

	blobmsg_add_u32(bb, "num_non1905_neighbors", p->non1905_nbr_num);
	non1905_array = blobmsg_open_array(bb, "non1905_neighbors");
	for (j = 0; j < p->non1905_nbr_num; j++) {
		hwaddr_ntoa(p->non1905_nbrlist[j], mac_str);
		blobmsg_add_string(bb, NULL, mac_str);
	}
	blobmsg_close_array(bb, non1905_array);

	if (is_self == 1)
		return;

	/*Here we dump the registar information*/
	for (i = 0; i < 3; i++) {
		if (g_registrar_freq_band[i] == 1) {
			if (flag == 1)
				strncpy(freq_str + strlen(freq_str), ",", 2);

			strncpy(freq_str + strlen(freq_str), get_registra_freq_string(i), 15);
			flag = 1;
		}
	}
	blobmsg_add_string(bb, "registrar_freq_band", freq_str);


	/*Here we are dumping the local interfaces of the 1905 neighbor*/
	/* and non1905 neighbor of the device*/
	if (p->non1905_nbr_num != 0) {
		hwaddr_ntoa(p->non1905_nbr_localintf, mac_str);
		blobmsg_add_string(bb, "non1905_nbr_localintf", mac_str);
	}

	if (p->nbr_num != 0) {
		hwaddr_ntoa(p->nbr_localintf, mac_str);
		blobmsg_add_string(bb, "1905_nbr_localintf", mac_str);
	}

	//Here we show the Device Local Interface Info
	blobmsg_add_u32(bb, "local_interface_nbr", p->local_intf_nbr);
	interfaces = blobmsg_open_array(bb, "interfaces");
	for (j = 0; j < p->local_intf_nbr; j++) {
		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->node_intf[j].interface_id, mac_str);
		blobmsg_add_string(bb, "interface_id", mac_str);
		blobmsg_add_string(bb, "media_type", get_interface_type_string(p->node_intf[j].media_type));
		blobmsg_add_string(bb, "power_status",
			p->node_intf[j].power_status ? "off" : "on");
		sprintf(oui, "%02x.%02x.%02x",
			p->node_intf[j].generic_phy_oui[0], p->node_intf[j].generic_phy_oui[1], p->node_intf[j].generic_phy_oui[2]);
		oui[9] = '\0';
		blobmsg_add_string(bb, "generic_phy_oui", oui);
		sprintf(phy_variant, "%02x", p->node_intf[j].generic_phy_variant);
		blobmsg_add_string(bb, "generic_phy_variant", phy_variant);
		blobmsg_add_string(bb, "generic_phy_url", p->node_intf[j].generic_phy_url);
		if (p->node_intf[j].media_specific_data_size == 10) {
			hwaddr_ntoa(p->node_intf[j].network_membership, mac_str);
			blobmsg_add_string(bb, "network_membership", mac_str);
			//blobmsg_add_string(bb, "role", get_role_type_string(p->node_intf[j].role));
			blobmsg_add_u32(bb, "ap_channel_band", p->node_intf[j].ap_channel_band);
			blobmsg_add_u32(bb, "freq_index1", p->node_intf[j].freq_index1);
			blobmsg_add_u32(bb, "freq_index2", p->node_intf[j].freq_index2);
		}
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, interfaces);

	//Here we show the Device L2 Nodes Info
	blobmsg_add_u32(bb, "l2_neighbor_num", p->l2_nbr_num);
	l2_neighbor = blobmsg_open_array(bb, "l2_neighbor");
	for (j = 0; j < p->l2_nbr_num; j++) {
		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->l2_nbr[j].l2_intf_addr, mac_str);
		blobmsg_add_string(bb, "local_intf_id", mac_str);
		//TODO add more information for l2
		hwaddr_ntoa(p->l2_nbr[j].l2_nbr_addr, mac_str);
		blobmsg_add_string(bb, "local_nbr_id", mac_str);
		for (k = 0; k < p->l2_nbr[j].behind_intf_nbr; k++) {
			size_t len, size;

			len = strlen(behind_mac_str);
			size = sizeof(behind_mac_str);

			hwaddr_ntoa(p->l2_nbr[j].behind_intf_addr[k], mac_str);
			if (k != 0)
				snprintf(behind_mac_str + len, size, "%s,", behind_mac_str);

			snprintf(behind_mac_str + len,  size, "%s", mac_str);
		}
		blobmsg_add_string(bb, "behind_mac_id", behind_mac_str);
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, l2_neighbor);

	//Here we show the Bridging tuple enteries
	blobmsg_add_u32(bb, "bridging_tuple_num", p->bridge_tuple_num);
	bridge_tuple = blobmsg_open_array(bb, "bridge_tuple");
	for (j = 0; j < p->bridge_tuple_num; j++) {
		table = blobmsg_open_table(bb, NULL);
		bridge_tuple_mac = blobmsg_open_array(bb, "br_mac");
		for (k = 0; k < p->br_tuple[j].tuple_num; k++) {
			hwaddr_ntoa(p->br_tuple[j].tuple_macid[k], mac_str);
			blobmsg_add_string(bb, "bridge_tuple_mac_id", mac_str);
		}
		blobmsg_close_array(bb, bridge_tuple_mac);
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, bridge_tuple);

	//Here we show the higher layer message data
	blobmsg_add_string(bb, "version", get_version(p->version));
	blobmsg_add_string(bb, "friendly_name", p->friendly_name);
	blobmsg_add_string(bb, "manufacturer_name", p->manufacturer_name);
	blobmsg_add_string(bb, "manufacturer_model", p->manufacturer_model);
	blobmsg_add_string(bb, "control_url", p->control_url);
	blobmsg_add_u8(bb, "is_easymesh", p->is_easymesh);

	//Here we show the ipv4 message structure
	blobmsg_add_u32(bb, "ipv4_itfr_num", p->ipv4_intf_nbr);
	ipv4_param = blobmsg_open_array(bb, "ipv4_params");

	for (j = 0; j < p->ipv4_intf_nbr; j++) {
#if 0
		void *service_arr = NULL;
#endif
		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->device_ipv4[j].mac_address, mac_str);
		blobmsg_add_string(bb, "mac_address", mac_str);
		blobmsg_add_string(bb, "ipv4_addr_type", get_ipv4_type(p->device_ipv4[j].ipv4_addr_type));

		// ipv4 address
		sprintf(addr, "%d.%d.%d.%d", p->device_ipv4[j].ipv4_addr[0],
				p->device_ipv4[j].ipv4_addr[1], p->device_ipv4[j].ipv4_addr[2],
				p->device_ipv4[j].ipv4_addr[3]);
		addr[15] = '\0';
		blobmsg_add_string(bb, "ipv4_address", addr);
		memset(addr, 0, 16);

		// dhcp address
		if (p->device_ipv4[j].ipv4_addr_type == IPV4_TYPE_DHCP) {
			sprintf(addr, "%d.%d.%d.%d", p->device_ipv4[j].dhcp_addr[0],
				p->device_ipv4[j].dhcp_addr[1],
				p->device_ipv4[j].dhcp_addr[2],
				p->device_ipv4[j].dhcp_addr[3]);
			addr[15] = '\0';
			blobmsg_add_string(bb, "dhcp_addr", addr);
		}

#if 0
		blobmsg_add_u16(bb, "mdns_no_services", p->device_ipv4->mdns_no_services);
		service_arr = blobmsg_open_array(bb, "mdns_services");

		for (k = 0; k < p->device_ipv4->mdns_no_services; k++) {
			int l;
			struct mdns_service_t *service;
			void *service_obj, *txt_arr;
			char date[20];

			service = p->device_ipv4->mdns_service[k];

			service_obj = blobmsg_open_table(bb, "NULL");
			blobmsg_add_string(bb, "application_protocol", service->application_protocol);
			blobmsg_add_string(bb, "transport_protocol", service->transport_protocol);
			blobmsg_add_string(bb, "domain", service->domain);
			blobmsg_add_u16(bb, "port", service->port);
			blobmsg_add_string(bb, "target", service->target);
			blobmsg_add_u16(bb, "txt_no_entries", service->txt_no_entries);

			txt_arr = blobmsg_open_array(bb, "text_records");
			for (l = 0; l < service->txt_no_entries; l++) {
				void *txt_table;

				txt_table = blobmsg_open_table(bb, "NULL");
				blobmsg_add_string(bb, "key", service->text_record[l]->key);
				blobmsg_add_string(bb, "value", service->text_record[l]->value);
				blobmsg_close_table(bb, txt_table);
			}
			blobmsg_close_table(bb, txt_arr);

			strftime(date, 20, "%Y-%m-%d %H:%M:%S", localtime(&service->time));
			blobmsg_add_string(bb, "last_update", date);
			blobmsg_add_u16(bb, "time_to_live", service->ttl);
			blobmsg_add_u16(bb, "priority", service->priority);
			blobmsg_add_u16(bb, "weight", service->weight);
			blobmsg_close_table(bb, service_obj);
		}
		blobmsg_close_array(bb, service_arr);
#endif

		blobmsg_close_table(bb, table);
	}

	blobmsg_close_array(bb, ipv4_param);

	//Here we show the ipv6 message structure
	blobmsg_add_u32(bb, "ipv6_itfr_num", p->ipv6_intf_nbr);
	ipv6_param = blobmsg_open_array(bb, "ipv6_params");
	for (j = 0; j < p->ipv6_intf_nbr; j++) {
		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->device_ipv6[j].mac_address, mac_str);
		blobmsg_add_string(bb, "mac_address", mac_str);
		blobmsg_add_string(bb, "ipv6_addr_type", get_ipv6_type(p->device_ipv6[j].ipv6_addr_type));
		//For ipv6 address
		sprintf(addr_v6, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
				p->device_ipv6[j].ipv6_addr[0], p->device_ipv6[j].ipv6_addr[1], p->device_ipv6[j].ipv6_addr[2],
				p->device_ipv6[j].ipv6_addr[3], p->device_ipv6[j].ipv6_addr[4], p->device_ipv6[j].ipv6_addr[5],
				p->device_ipv6[j].ipv6_addr[6], p->device_ipv6[j].ipv6_addr[7], p->device_ipv6[j].ipv6_addr[8],
				p->device_ipv6[j].ipv6_addr[9], p->device_ipv6[j].ipv6_addr[10], p->device_ipv6[j].ipv6_addr[11],
				p->device_ipv6[j].ipv6_addr[12], p->device_ipv6[j].ipv6_addr[13], p->device_ipv6[j].ipv6_addr[14],
				p->device_ipv6[j].ipv6_addr[15]);
		addr_v6[39] = '\0';
		blobmsg_add_string(bb, "ipv6_address", addr_v6);
		// origin address
		if (p->device_ipv6[j].ipv6_addr_type == IPV6_TYPE_SLAAC) {
			memset(addr_v6, 0, 40);
			sprintf(addr_v6, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
					p->device_ipv6[j].ipv6addr_origin[0], p->device_ipv6[j].ipv6addr_origin[1],
					p->device_ipv6[j].ipv6addr_origin[2], p->device_ipv6[j].ipv6addr_origin[3],
					p->device_ipv6[j].ipv6addr_origin[4], p->device_ipv6[j].ipv6addr_origin[5],
					p->device_ipv6[j].ipv6addr_origin[6], p->device_ipv6[j].ipv6addr_origin[7],
					p->device_ipv6[j].ipv6addr_origin[8], p->device_ipv6[j].ipv6addr_origin[9],
					p->device_ipv6[j].ipv6addr_origin[10], p->device_ipv6[j].ipv6addr_origin[11],
					p->device_ipv6[j].ipv6addr_origin[12], p->device_ipv6[j].ipv6addr_origin[13],
					p->device_ipv6[j].ipv6addr_origin[14], p->device_ipv6[j].ipv6addr_origin[15]);
			addr_v6[39] = '\0';
			blobmsg_add_string(bb, "origin", addr_v6);
		}
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, ipv6_param);

}

void topologyd_dump_node_nbr(struct topologyd_private *priv, struct blob_buf *b)
{
	char mac_str[18] = { 0x0 };
	void *nodes_array;
	struct node *p;
	int i;

	/* Add details of own node first */
	p = &priv->selfnode;

	hwaddr_ntoa(p->hwaddr, mac_str);
	blobmsg_add_string(b, "ieee1905_macaddr", mac_str);

	nodes_array = blobmsg_open_array(b, "neighbors");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;

			hwaddr_ntoa(p->hwaddr, mac_str);
			blobmsg_add_string(b, NULL, mac_str);
		}
	}
	blobmsg_close_array(b, nodes_array);
}

void topologyd_dump_tree(struct topologyd_private *priv, struct blob_buf *bb)
{
	void *nodes_array, *table;
	struct node *p;
	int i;

	/* Add details of own node first */
	p = &priv->selfnode;

	table = blobmsg_open_table(bb, "self");
	topologyd_dump_node(bb, p, 1);
	blobmsg_close_table(bb, table);

	nodes_array = blobmsg_open_array(bb, "nodes");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

			if (hwaddr_is_zero(p->hwaddr))	// TODO: unlikely..
				continue;

			table = blobmsg_open_table(bb, NULL);

			topologyd_dump_node(bb, p, 0);
			blobmsg_close_table(bb, table);
		}
	}
	blobmsg_close_array(bb, nodes_array);
}

static int topologyd_dump(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	topologyd_dump_tree(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int topologyd_nodes(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
		container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	topologyd_dump_node_nbr(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}


void topologyd_print_changelog(struct blob_buf *bb, struct topologyd_private *priv, uint32_t start, uint32_t end)
{
	struct topology_changelog *log;
	char rpt_ifmacstr[18] = { 0 };
	char rpt_macstr[18] = { 0 };
	char nbr_macstr[18] = { 0 };
	void *t;
	int i;
	struct tm *info;
	time_t tmp_t;
	char str_tm[20];

	dbg("changelog start value is [%d] end value is [%d]", start, end);
	if (bb == NULL || priv == NULL)
		return;

	for (i = start ; i <= end ; i++) {
		log = &priv->topo.changelog[i];
		t = blobmsg_open_table(bb, NULL);

		hwaddr_ntoa(log->rpt_macaddr, rpt_macstr);
		blobmsg_add_string(bb, "reporter", rpt_macstr);

		hwaddr_ntoa(log->rpt_ifmacaddr, rpt_ifmacstr);
		blobmsg_add_string(bb, "reporter_interface", rpt_ifmacstr);

		hwaddr_ntoa(log->nbr_macaddr, nbr_macstr);
		blobmsg_add_string(bb, "neighbor", nbr_macstr);

		blobmsg_add_string(bb, "is1905_neighbor",
			log->is1905_nbr ? "IEEE1905" : "Non-IEEE1905");

		blobmsg_add_string(bb, "event_type",
			log->type ? "delete" : "add");

		tmp_t = log->timestamp;
		info = localtime(&tmp_t);
		strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%S", info);
		blobmsg_add_string(bb, "timestamp", str_tm);
		blobmsg_close_table(bb, t);
	}
}
void topologyd_dump_changelog(struct topologyd_private *priv, struct blob_buf *bb)
{
	void *a;

	blobmsg_add_u32(bb, "num_changelog", priv->topo.num_changelog);

	a = blobmsg_open_array(bb, "changelog");
	if (priv->topo.changelog_front == -1) {
		blobmsg_close_array(bb, a);
		return;
	}
	if (priv->topo.changelog_front > priv->topo.changelog_rear) {
		topologyd_print_changelog(bb, priv, priv->topo.changelog_front, topology_log_max-1);
		topologyd_print_changelog(bb, priv, 0, priv->topo.changelog_rear);
	} else {
		topologyd_print_changelog(bb, priv, 0, priv->topo.changelog_rear);
	}

	blobmsg_close_array(bb, a);
}

static int topologyd_changelog(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	topologyd_dump_changelog(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int topologyd_status(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "status", priv->status == AVAILABLE ?
					"available" : "incomplete");

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static void topologyd_periodic_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, refresh_timer);

	dbg("Refresh topology...\n");

	topologyd_algo_run(priv);

	if (priv->config.refresh_int)
		uloop_timeout_set(&priv->refresh_timer, priv->config.refresh_int * 1000);
}

static void topologyd_refresh_notify(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, topo_notify_timer);

	dbg("Refresh Notify topology...\n");

	topologyd_algo_run(priv);

}

static void topologyd_change_status(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, status_timer);

	dbg("Change the status of topology...\n");
	priv->status = 1;
}


static void topologyd_start_heartbeat(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
			container_of(t, struct topologyd_private, heartbeat);
	sigset_t waiting_mask;

	sigpending(&waiting_mask);

	if (sigismember(&waiting_mask, SIGINT)) {
		dbg("|%s:%d| Received SIGINT, terminating\n", __func__, __LINE__);
		exit(0);
	}

	if (sigismember(&waiting_mask, SIGTERM)) {
		dbg("|%s:%d| Received SIGSTP, terminating\n", __func__, __LINE__);
		exit(0);
	}

	if (sigismember(&waiting_mask, SIGHUP)) {

		dbg("|%s:%d| Received SIGHUP, reload network config\n", __func__, __LINE__);

		signal(SIGHUP, SIG_IGN);
		//Here reload the network interface in the hosts
		host_get_network_dump(priv);
	}
	uloop_timeout_set(&priv->heartbeat, 1 * 1000);
}

static int topologyd_run(struct topologyd_private *priv)
{
	if (priv->use_ieee1905) {
		int ret;

		//Here we fill the albridge mac address that is needed to be ignored in
		//host.devices
		topologyd_get_albridge_macaddr(priv);
		ret = topologyd_get_ieee1905_macaddr(priv);
		ret |= topologyd_send_ieee1905_topology_query(priv, NULL);

		if (ret) {
			err("Failed to run topologyd...\n");
			return -1;
		}
		topologyd_periodic_refresh(&priv->refresh_timer);
	}

	topologyd_start_heartbeat(&priv->heartbeat);
	host_periodic_refresh(&priv->host_refresh_timer);

	return 0;
}

static uint32_t topologyd_get_ieee1905_object(struct topologyd_private *priv)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(priv->ctx, IEEE1905_OBJECT, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", IEEE1905_OBJECT);
		return OBJECT_INVALID;
	}

	return id;
}

int topologyd_publish_object(struct topologyd_private *priv, const char *objname)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct ubus_method m[6] = {
		UBUS_METHOD_NOARG("refresh", topologyd_refresh),
		UBUS_METHOD_NOARG("status", topologyd_status),
		UBUS_METHOD_NOARG("dump", topologyd_dump),
		UBUS_METHOD_NOARG("changelog", topologyd_changelog),
		UBUS_METHOD_NOARG("nodes", topologyd_nodes),
		UBUS_METHOD_NOARG("hosts", host_nodes),
	};
	int num_methods = ARRAY_SIZE(m);
	int ret;


	obj = &priv->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = objname;
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(priv->ctx, obj);
	if (ret) {
		fprintf(stderr, "Failed to add 'obj' err = %s\n",
							ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	fprintf(stderr, "Added object '%s'\n", objname);

	return 0;
}

int topologyd_start(void)
{
	struct topologyd_private *priv;
	int ret = 0;

	sigset_t base_mask;

	sigemptyset(&base_mask);
	sigaddset(&base_mask, SIGINT);
	sigaddset(&base_mask, SIGTERM);
	sigaddset(&base_mask, SIGHUP);

	sigprocmask(SIG_SETMASK, &base_mask, NULL);
	set_sighandler(SIGPIPE, SIG_IGN);

	priv = calloc(1, sizeof(*priv));
	if (!priv)
		return -ENOMEM;


	info("Starting topologyd...");

	priv->use_ieee1905 = topologyd_use_ieee1905();

	t_cmdu_ackq_init(&priv->cmdu_ack_q);

	uloop_init();
	priv->ctx = ubus_connect(ubus_socket);
	if (!priv->ctx) {
		err("Failed to connect to ubus\n");
		free(priv);
		return -1;
	}

	ubus_add_uloop(priv->ctx);

	topologyd_config_defaults(&priv->config);

#ifdef HOST_WAN_STATS
	/*host wan stats collection thread creation*/
	host_node_create_rd_wr_sockets(priv);
	host_wan_stats_thread(priv);
#endif

	host_config_defaults(&priv->host_global_cfg);
	topologyd_config_user(&priv->config);
	host_config_global(&priv->host_global_cfg);
	host_create_file();

	//Here start the host refresh timer
	priv->host_refresh_timer.cb = host_periodic_refresh;

	/*Here we set the cb for the arp table read timer*/
	priv->host_arp_timer.cb = host_arp_table_refresh;
	priv->host_wifi_timer.cb = host_wifi_table_refresh;
	priv->heartbeat.cb = topologyd_start_heartbeat;

	//Here add the event hadler for the host
	//wifi connection events
	if (!(priv->is_mapagent)) {
		priv->wifi_evh.cb = host_event_handler;
		ubus_register_event_handler(priv->ctx, &priv->wifi_evh, WIFI_EVENTS);
	}


	if (priv->use_ieee1905) {

		for (;;) {
			priv->ieee1905 = topologyd_get_ieee1905_object(priv);
			if (priv->ieee1905 != OBJECT_INVALID)
				break;

			sleep(2);
		}

		//Here we needto register for event so that resubscribe can be done
		//after autoconfiguration
		priv->obj_register.cb = uobj_add_event_handler;
		ubus_register_event_handler(priv->ctx, &priv->obj_register, "ubus.object.add");

		priv->is_mapagent = is_local_mapagent_available();
		dbg("mapagent is present = [%d]\n", priv->is_mapagent);


		priv->topo_notify_timer.cb = topologyd_refresh_notify;
		priv->refresh_timer.cb = topologyd_periodic_refresh;
		priv->topo.changelog = (struct topology_changelog *) calloc
			((priv->config.maxlog), sizeof(struct topology_changelog));
		if (priv->topo.changelog == NULL)
			goto out_and_exit;

		topology_log_max = priv->config.maxlog;
		dbg("Dynamic array size %d", priv->config.maxlog);
		priv->topo.changelog_front = priv->topo.changelog_rear = -1;

		priv->status = 0;
		priv->status_timer.cb = topologyd_change_status;
	} else {

		ret = register_nlevents(priv);
		if (ret) {
			free(priv);
			return -1;
		}

		host_get_neighbors(priv);
	}

	/*Here we need to fill the host structures*/
	config_init_host_data("hosts", "host", priv);
	host_get_network_dump(priv);
	if (!priv->config.enabled) {
		info("topologyd not enabled.\n");
	} else {
		topologyd_publish_object(priv, TOPOLOGY_OBJECT);
		topologyd_run(priv);
	}
#ifdef HOST_WAN_STATS
	host_wan_stats_fd_register(priv);
	host_wan_stats_collect_timer(&priv->host_wan_stats_timer);
#endif
	topology_subscribe_for_cmdus(priv);

	uloop_run();

	//ubus_unregister_event_handler(priv->ctx, &priv->ev);
	//topologyd_remove_object(priv->ctx);   //TODO
	if (priv->wifi_stations)
		json_object_put(priv->wifi_stations);
	if (priv->network_dump)
		json_object_put(priv->network_dump);
out_and_exit:
	unregister_nlevents();
	ubus_free(priv->ctx);
	uloop_done();
	t_cmdu_ackq_free(&priv->cmdu_ack_q);
	free(priv);

	return 0;
}

void topologyd_stop(struct topologyd_private *priv)
{
	if (!priv)
		exit(0);

	//ubus_unregister_event_handler(priv->ctx, &priv->ev);
	//topologyd_remove_object(priv->ctx);
	ubus_free(priv->ctx);
	uloop_done();
	t_cmdu_ackq_free(&priv->cmdu_ack_q);
	free(priv);
}

// Check if the changelog queue is full
int is_full(struct topology *t)
{
	if ((t->changelog_front == t->changelog_rear + 1) || (t->changelog_front == 0 && t->changelog_rear == topology_log_max - 1))
		return 1;
	return 0;
}

// Adding an entry in changelog
void enqueue_changelog(struct topology *t, struct topology_changelog *c)
{
	dbg("Inside %s\n", __func__);
	struct topology_changelog *elem;

	if (is_full(t)) {
		dbg("Changelog queue is full deleting the front node\n");
		t->changelog_front = (t->changelog_front + 1) % topology_log_max;
		t->num_changelog--;
	}

	if (t->changelog_front == -1)
		t->changelog_front = 0;
	t->changelog_rear = (t->changelog_rear + 1) % topology_log_max;
	elem = &(t->changelog[t->changelog_rear]);
	memset(elem, 0, sizeof(struct topology_changelog));
	memcpy(elem->rpt_macaddr, c->rpt_macaddr, 6);
	memcpy(elem->rpt_ifmacaddr, c->rpt_ifmacaddr, 6);
	memcpy(elem->nbr_macaddr, c->nbr_macaddr, 6);
	elem->is1905_nbr = c->is1905_nbr;
	elem->type = c->type;
	elem->timestamp = time(NULL);
	//Maintain the number of elements in changelog
	t->num_changelog++;
}

void changelog_copy_node_info(struct topologyd_private *t, struct node *p, int32_t event, bool is1905)
{
	dbg("Inside %s...\n", __func__);
	struct topology_changelog elem = { 0 };
	struct host_node *ret = NULL;

	if (t == NULL || p == NULL)
		return;

	if (event == 0) {
		memcpy(elem.rpt_macaddr, p->hwaddr, 6);
		memcpy(elem.rpt_ifmacaddr, p->nbr_localintf, 6);
	} else {
		memcpy(elem.rpt_macaddr, t->ieee1905_macaddr, 6);
		memcpy(elem.rpt_ifmacaddr, t->selfnode.nbr_localintf, 6);
	}

	memcpy(elem.nbr_macaddr, p->hwaddr, 6);
	elem.is1905_nbr = is1905;
	elem.type = event;
	enqueue_changelog(&(t->topo), &elem);

	//Send event
	topology_send_node_event(&elem, p->ingress_ifr_name);

	//Here we add or del the node in the host
	if (event == 0) {
		struct host_node *hn;
		hn = host_node_lookup(t->host.node_htable, p->hwaddr);
		if (!hn) {
			hn = host_topo_node_add(t, p, NULL, 0);
			if (!hn) {
				err("Failed to add node in the hosts\n");
				return;
			}
			if (hn->is_ipaddr)
				host_send_client_event(t, hn, 1);
		}
	} else if (event == 1) {
		ret = host_topo_node_del(t, p, NULL, 0);
		if (!ret)
			err("Failed to delete node in the hosts\n");
	}
}

static void _update_changelog_info(struct topologyd_private *priv,
				   struct node *dest, struct node *src,
				   int change_type)
{
	struct topology_changelog elem;
	struct host_node *hn;
	int found;
	int i, j;


	dbg("%s() ===>\n", __func__);
	for (i = 0; i < dest->non1905_nbr_num; i++) {
		found = 0;
		for (j = 0; j < src->non1905_nbr_num; j++) {
			if (!memcmp(dest->non1905_nbrlist[i], src->non1905_nbrlist[j], 6)) {
				found = 1;
				break;
			}
		}

		if (found == 0) {
			memcpy(elem.nbr_macaddr, dest->non1905_nbrlist[i], 6);
			memcpy(elem.rpt_macaddr, dest->hwaddr, 6);
			memcpy(elem.rpt_ifmacaddr, dest->non1905_nbr_localintf, 6);
			elem.is1905_nbr = 0;
			elem.type = change_type;
			enqueue_changelog(&(priv->topo), &elem);

			//Send event
			topology_send_node_event(&elem, src->ingress_ifr_name);

			dbg("%s %d\n", __func__, __LINE__);

			/* Here we need to change the status of node in hosts */
			hn = host_node_lookup(priv->host.node_htable, elem.nbr_macaddr);
			if (hn && hn->intf_type != HOST_TYPE_WIFI) {
				hn = host_topo_node_del(priv, NULL, elem.nbr_macaddr, 0);
				if (!hn)
					err("Failed to del node in the hosts\n");
			} else {
				//XXX: if (change_type == 0) needed?
				hn = host_topo_node_add(priv, NULL, elem.nbr_macaddr, 0);
				if (!hn) {
					err("Failed to add non1905 node in the hosts\n");
					continue;
				}
				if (hn->is_ipaddr)
					host_send_client_event(priv, hn, 1);
			}
		}
	}
}

void update_changelog_info(struct topologyd_private *priv, struct node *dest, struct node *src)
{
	dbg("Inside %s..\n", __func__);

	if (dest == NULL || src == NULL || priv == NULL)
		return;

	dbg("The dest num [%d] src num [%d]", dest->non1905_nbr_num,
	    src->non1905_nbr_num);

	/* Check the update in non1905 neighbor node */
	if (dest->non1905_nbr_num != src->non1905_nbr_num) {
		if (dest->non1905_nbr_num > src->non1905_nbr_num)
			_update_changelog_info(priv, dest, src, NBR_NODE_DEL);
		else
			_update_changelog_info(priv, src, dest, NBR_NODE_ADD);
	} else {
		dbg("%s %d\n", __func__, __LINE__);
		_update_changelog_info(priv, dest, src, NBR_NODE_DEL);
		_update_changelog_info(priv, src, dest, NBR_NODE_ADD);
	}
}

struct cmdu_buff *topo_gen_query(uint16_t type)
{
	struct cmdu_buff *resp;
	int ret;
	uint16_t mid = 0;
	struct tlv_linkmetric_query *lq;
	struct tlv *t;

	dbg("Inside %s...\n", __func__);
	resp = t_cmdu_alloc_simple(type, &mid);
	if (!resp) {
		err("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	dbg(" function %s mid = [%hu]\n", __func__, t_cmdu_get_mid(resp));

	if (type == CMDU_TYPE_LINK_METRIC_QUERY) {
		t = t_cmdu_reserve_tlv(resp, 256);
		if (!t) {
			t_cmdu_free(resp);
			return NULL;
		}

		t->type = TLV_TYPE_LINK_METRIC_QUERY;
		t->len = sizeof(*lq);
		lq = (struct tlv_linkmetric_query *)t->data;
		lq->nbr_type = LINKMETRIC_QUERY_NEIGHBOR_ALL;
		lq->query_type = LINKMETRIC_QUERY_TYPE_BOTH;

		ret = t_cmdu_put_tlv(resp, t);
		if (ret) {
			fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
			t_cmdu_free(resp);
			return NULL;
		}
	}

	t_cmdu_put_eom(resp);
	return resp;
}

int topologyd_send_ieee1905_query(struct topologyd_private *p,
	uint8_t *dest, uint16_t type, uint8_t *data)
{
	struct cmdu_buff *cmdu;
	uint16_t ret = 0;

	dbg("Inside %s...\n", __func__);
	if (dest == NULL) {
		send_buildcmdu(p, type);
	} else {
	dbg("Inside %s...dest = ["MACFMT"]\n", __func__, MAC2STR(dest));
		cmdu = topo_gen_query(type);
		if (!cmdu)
			return -1;
		ret = send_cmdu(p, cmdu, dest);
		t_cmdu_free(cmdu);
	}
	return ret;
}

const char *get_interface_type_string(uint16_t if_type)
{

	switch (if_type) {
	case I1905_802_3U_FAST_ETHERNET:
		return "802_3U_FAST_ETHERNET";
	case I1905_802_3AB_GIGABIT_ETHERNET:
		return "802_3AB_GIGABIT_ETHERNET";
	case I1905_802_11B_2_4_GHZ:
		return "802_11B_2_4_GHZ";
	case I1905_802_11G_2_4_GHZ:
		return "802_11G_2_4_GHZ";
	case I1905_802_11A_5_GHZ:
		return "802_11A_5_GHZ";
	case I1905_802_11N_2_4_GHZ:
		return "802_11N_2_4_GHZ";
	case I1905_802_11N_5_GHZ:
		return "802_11N_5_GHZ";
	case I1905_802_11AC_5_GHZ:
		return "802_11AC_5_GHZ";
	case I1905_802_11AD_60_GHZ:
		return "802_11AD_60_GHZ";
	case I1905_802_11AF_GHZ:
		return "802_11AF_GHZ";
	case I1905_802_11AX_2_4_GHZ:
		return "802_11AX_2_4_GHZ";
	case I1905_802_11AX_5_GHZ:
		return "802_11AX_5_GHZ";
	case I1905_1901_WAVELET:
		return "1901_WAVELET";
	case I1905_1901_FFT:
		return "1901_FFT";
	case I1905_MOCA_V1_1:
		return "MOCA_V1_1";
	}
	return"UNKNOWN";
};

#if 0
const char *get_interface_type_string(uint16_t if_type)
{

	switch (if_type) {
	case MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET:
		return "IEEE 802.3u";
	case MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET:
		return "IEEE 802.3ab";
	case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
		return "IEEE 802.11b";
	case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
		return "IEEE 802.11g";
	case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
		return "IEEE 802.11a";
	case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		return "IEEE 802.11n 2.4";
	case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
		return "IEEE 802.11n 5.0";
	case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		return "IEEE 802.11ac";
	case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		return "IEEE 802.11ad";
	case MEDIA_TYPE_IEEE_802_11AF_GHZ:
		return "IEEE 802.11af";
	case MEDIA_TYPE_IEEE_1901_WAVELET:
		return "IEEE 1901 Wavelet";
	case MEDIA_TYPE_IEEE_1901_FFT:
		return "IEEE 1901 FFT";
	case MEDIA_TYPE_MOCA_V1_1:
		return "MoCAv1.1";
	case MEDIA_TYPE_UNKNOWN:
		return "Generic PHY";
	}
	return "UNKNOWN";
}
#endif
#if 0
const char *get_role_type_string(uint16_t role_type)
{
	switch (role_type) {
	case IEEE80211_SPECIFIC_INFO_ROLE_AP:
		return"AP";
	case IEEE80211_SPECIFIC_INFO_ROLE_NON_AP_NON_PCP_STA:
		return"non-AP/non-PCP STA";
	case IEEE80211_SPECIFIC_INFO_ROLE_WIFI_P2P_CLIENT:
		return"Wi-Fi P2P Client";
	case IEEE80211_SPECIFIC_INFO_ROLE_WIFI_P2P_GROUP_OWNER:
		return"Wi-Fi P2P Group Owner";
	case IEEE80211_SPECIFIC_INFO_ROLE_AD_PCP:
		return"802.11adPCP";
	}
	return "UNKNOWN";
}
#endif
const char *get_registra_freq_string(uint32_t band)
{
	switch (band) {
	case 0:
		return"802.11 2.4 GHz";
	case 1:
		return"802.11 5 GHz";
	case 2:
		return"802.11 60 GHz";
	}
	return "UNKNOWN";
}

const char *get_version(uint8_t ver)
{
	switch (ver) {
	case PROFILE_1905_1:
		return"1905.1";
	case PROFILE_1905_1A:
		return"1905.1a";
	}
	return "UNKNOWN";
}

const char *get_ipv4_type(uint8_t ipv4_type)
{
	switch (ipv4_type) {
	case IPV4_TYPE_UNKNOWN:
		return"IPV4_UNKNOWN";
	case IPV4_TYPE_DHCP:
		return"IPV4_DHCP";
	case IPV4_TYPE_STATIC:
		return"IPV4_STATIC_IP";
	case IPV4_TYPE_AUTOIP:
		return"IPV4_AUTO_IP";
	}
	return "UNKNOWN";
}

const char *get_ipv6_type(uint8_t ipv6_type)
{
	switch (ipv6_type) {
	case IP6_TYPE_UNKNOWN:
		return"IPV6_UNKNOWN";
	case IP6_TYPE_LINKLOCAL:
		return"IPV6_LINKLOCAL";
	case IP6_TYPE_DHCP:
		return"IPV6_DHCP";
	case IP6_TYPE_STATIC:
		return "IPV6_STATIC_IP";
	case IP6_TYPE_SLAAC:
		return"IPV6_SLAAC";
	}
	return "UNKNOWN";
}

void topology_send_node_event(struct topology_changelog *c, char *ingress)
{

	dbg("Inside %s...\n", __func__);
	struct ubus_context *sctx = ubus_connect(NULL);

	if (!sctx) {
		err("Failed to connect with ubus");
		return;
	}

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (c) {
		char mac[18] = {0}, src_mac[18] = {0};

		hwaddr_ntoa(c->nbr_macaddr, mac);
		blobmsg_add_string(&bb, "macaddr", mac);
		hwaddr_ntoa(c->rpt_macaddr, src_mac);
		blobmsg_add_string(&bb, "src_macaddr", src_mac);
		if (c->is1905_nbr == 1)
			blobmsg_add_u8(&bb, "is1905", true);
		else
			blobmsg_add_u8(&bb, "is1905", false);


	}
	if (ingress)
		blobmsg_add_string(&bb, "ingress", ingress);


	//Here we need to check if this is a add event or a delete event
	if (c->type == 0)
		ubus_send_event(sctx, TOPOLOGY_EVENT_ADD, bb.head);
	else
		ubus_send_event(sctx, TOPOLOGY_EVENT_DEL, bb.head);

	blob_buf_free(&bb);

	ubus_free(sctx);
}

void topologyd_update_changelog_firstentry(struct topologyd_private *priv, struct node *dest)
{
	dbg("Inside %s..\n", __func__);
	int i = 0;
	struct topology_changelog elem;
	uint8_t node_mac[6];
	struct host_node *hn = NULL;

	if (dest == NULL || priv == NULL)
		return;
	dbg("dest num for non1905 nodes [%d]", dest->non1905_nbr_num);
	//Check the update in non neighbor node
	if (dest->non1905_nbr_num > 0) {
		//Here add non ieee1905 neighbor
		for (i = 0; i < dest->non1905_nbr_num ; i++) {
			memcpy(elem.nbr_macaddr, dest->non1905_nbrlist[i], 6);
			memcpy(node_mac, dest->non1905_nbrlist[i], 6);
			memcpy(elem.rpt_macaddr, dest->hwaddr, 6);
			memcpy(elem.rpt_ifmacaddr, dest->non1905_nbr_localintf, 6);
			elem.is1905_nbr = 0;
			elem.type = 0;
			enqueue_changelog(&(priv->topo), &elem);
			//Send event
			topology_send_node_event(&elem, dest->ingress_ifr_name);
			/*Here we need to add the node in the host*/
			hn = host_node_lookup(priv->host.node_htable, node_mac);
			if (!hn) {
				hn = host_topo_node_add(priv, NULL, node_mac, 0);
				if (!hn) {
					err("Failed to add node in the hosts\n");
					continue;
				}
				if (hn->is_ipaddr)
					host_send_client_event(priv, hn, 1);
			}
		}
	}
}

static void send_cmdu_cb(struct ubus_request *req,
		int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	dbg("Inside %s...\n", __func__);

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);

	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp)) {
		*mid = json_object_get_int(tmp);
		fprintf(stdout, "%s:%d agent map-mid:%d\n", __func__, __LINE__, *mid);
	}

	json_object_put(jobj);
}

int get_resp_type(uint16_t *resp_type, uint16_t msg_type)
{

	switch (msg_type) {
	case CMDU_TYPE_LINK_METRIC_QUERY:
		*resp_type = CMDU_TYPE_LINK_METRIC_RESPONSE;
	break;
	case CMDU_TYPE_TOPOLOGY_QUERY:
		*resp_type = CMDU_TYPE_TOPOLOGY_RESPONSE;
	break;
	case CMDU_TYPE_HIGHER_LAYER_QUERY:
		*resp_type = CMDU_TYPE_HIGHER_LAYER_RESPONSE;
	break;
	default:
		return 1;
	}
	return 0;
}

uint16_t send_cmdu(struct topologyd_private *priv, struct cmdu_buff *cmdu, uint8_t *dst)
{
	struct blob_buf b = { 0 };
	char dst_addr[18] = { 0 };
	uint16_t msgid = 0;
	int ret = 0;
	uint32_t id;
	uint16_t resp_type;
	uint16_t msg_type;

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	msg_type = t_cmdu_get_type(cmdu);
	blobmsg_add_u32(&b, "type", msg_type);
	memcpy(cmdu->origin, priv->selfnode.hwaddr, 6);

	if (dst == NULL)
		hwaddr_ntoa(priv->selfnode.hwaddr, dst_addr);
	else
		hwaddr_ntoa(dst, dst_addr);
	blobmsg_add_string(&b, "dst", dst_addr);

	blobmsg_add_u32(&b, "mid", (uint32_t)t_cmdu_get_mid(cmdu));

	if (cmdu->datalen) {
		char *tlv_str = NULL;
		uint16_t len = 0;

		len = (cmdu->datalen * 2) + 1;

		tlv_str = calloc(len, sizeof(char));
		if (!tlv_str)
			goto out;

		btostr(cmdu->data, cmdu->datalen, tlv_str);
		tlv_str[len-1] = '\0';

		blobmsg_add_string(&b, "data", tlv_str);
		free(tlv_str);
	} else
		blobmsg_add_string(&b, "data", "000000");

	dbg("|%s:%d|cmdu:%u|egress:%s|dst:%s|mid:%u|fid:%u\n", __func__, __LINE__,
		t_cmdu_get_type(cmdu),
		"TODO", dst_addr, t_cmdu_get_mid(cmdu), t_cmdu_get_fid(cmdu));

	if (ubus_lookup_id(priv->ctx, I1905_OBJ, &id)) {
		dbg("[%s:%d] not present i1905", __func__, __LINE__);
		goto out;
	}

	ret = ubus_invoke(priv->ctx, id, "cmdu",
				b.head, send_cmdu_cb,
				&msgid,
				1000);
	if (ret) {
		dbg("[%s:%d] ubus call failed for |i1905 send|",
			 __func__, __LINE__);
		goto out;
	}
	dbg("|%s:%d| msgid = %d\n", __func__, __LINE__, msgid);

	/* Store mid of the sent CMDU with an expected response type.
	 * On receiving a 'resp_type' cmdu, lookup in the ackq and
	 * drop or handle it.
	 */
	ret = get_resp_type(&resp_type, msg_type);

	if (ret != 0)
		goto out;
	t_cmdu_ackq_enqueue(&priv->cmdu_ack_q, resp_type, msgid,
			  dst, 60000, 0, NULL);
	dbg("Inside %s...ENQUEUE msg_type = [%d] msg_mid = [%d] src =["MACFMT"]\n", __func__,
		resp_type, msgid, MAC2STR(dst));
	ret = msgid;
out:
	blob_buf_free(&b);

	return ret;
}

static void send_buildcmdu_cb(struct ubus_request *req,
		int type, struct blob_attr *msg)
{
	json_object *jroot = NULL;
	char *str;
	char *tlv_str;
	int len;
	uint16_t msg_type = 0;
	uint8_t *tlv = NULL;
	uint32_t b_len = 0;
	struct cmdu_buff *cstruct = NULL;
	struct topologyd_private *priv;
	uint16_t mid;

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	if (!msg)
		return;

	priv  = (struct topologyd_private *)req->priv;
	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		msg_type = _json_obj_get_int(jroot, "type");
		tlv_str = _json_obj_get_string(jroot, "data");
		if (tlv_str == NULL)
			return;
		len = strlen(tlv_str);
		dbg(" msg type is %d The TLV string is %s LENGTH Id %d\n", msg_type, tlv_str, len);
		b_len = (len/2);
		dbg("The binary length of the tlv is %d", b_len);
		tlv = (uint8_t *) malloc((b_len) * sizeof(uint8_t));
		if (tlv == NULL) {
			err("No Memory\n");
			if (tlv_str != NULL)
				free(tlv_str);
			return;
		}
		strtob(tlv_str, b_len, tlv);
		if (tlv_str != NULL)
			free(tlv_str);
		json_safe_put(jroot);
	}
	cstruct = t_cmdu_alloc_custom(msg_type, &mid, NULL, priv->selfnode.hwaddr, tlv, b_len);
	if (cstruct == NULL) {
		if (tlv != NULL)
			free(tlv);
		return;
	}

	if (msg_type == CMDU_TYPE_TOPOLOGY_RESPONSE) {
		topologyd_process_topology_response(cstruct, priv, NULL);
	} else if (msg_type == CMDU_TYPE_HIGHER_LAYER_RESPONSE) {
		topologyd_process_higherlayer_response(cstruct, priv);
	} else if (msg_type == CMDU_TYPE_LINK_METRIC_RESPONSE) {
		topologyd_process_linkmetric_response(cstruct, priv);
	}

	dbg("|%s:%d| Entry\n", __func__, __LINE__);
	if (tlv != NULL)
		free(tlv);
	t_cmdu_free(cstruct);
	return;
}

int send_buildcmdu(struct topologyd_private *priv, uint16_t msg_type)
{
	struct blob_buf b = { 0 };
	int ret = 0;
	uint32_t id;

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	blobmsg_add_u32(&b, "type", (uint32_t)msg_type);

	if (ubus_lookup_id(priv->ctx, I1905_OBJ, &id)) {
		dbg("[%s:%d] not present i1905", __func__, __LINE__);
		goto out;
	}

	ret = ubus_invoke(priv->ctx, id, "buildcmdu",
			b.head, send_buildcmdu_cb,
			priv,
			5000);
	if (ret) {
		dbg("[%s:%d] ubus call failed for |i1905 buildcmdu|",
			 __func__, __LINE__);
		goto out;
	}
out:
	blob_buf_free(&b);
	return ret;
}

struct cmdu_buff *topo_gen_topology_query(struct topologyd_private *priv, uint8_t profile)
{
	struct cmdu_buff *resp;
	struct tlv *t;
	struct tlv_map_profile *data;
	uint16_t mid = 0;
	int ret = 0;

	dbg("Inside %s...\n", __func__);
	resp = t_cmdu_alloc_simple(CMDU_TYPE_TOPOLOGY_QUERY, &mid);
	if (!resp) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}

	t = t_cmdu_reserve_tlv(resp, 20);
	if (!t)
		goto out;

	t->type = MAP_TLV_MULTIAP_PROFILE;
	t->len = sizeof(*data);
	data = (struct tlv_map_profile *)t->data;
	data->profile = profile;

	ret = t_cmdu_put_tlv(resp, t);
	if (ret) {
		err("%s: error: cmdu_put_tlv()\n", __func__);
		goto out;
	}

	dbg(" function %s mid = [%hu]\n", __func__, t_cmdu_get_mid(resp));

	t_cmdu_put_eom(resp);
	return resp;
out:
	t_cmdu_free(resp);
	return NULL;
}

int topologyd_send_ieee1905_topology_query(struct topologyd_private *p,
		uint8_t *dest)
{
	struct cmdu_buff *cmdu;

	dbg("Inside %s...\n", __func__);
	if (dest == NULL)
		send_buildcmdu(p, CMDU_TYPE_TOPOLOGY_RESPONSE);
	else {
		cmdu = topo_gen_topology_query(p, p->config.profile);
		if (!cmdu)
			return -1;
		send_cmdu(p, cmdu, dest);
		t_cmdu_free(cmdu);
	}
	return 0;
}

void topologyd_process_topology_notification(struct cmdu_buff *cstruct,
						struct topologyd_private *priv)
{
	struct tlv *tv[1][TLV_MAXNUM] = {0};
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_CLIENT_ASSOCIATION_EVENT, .present = TLV_PRESENT_ONE },
	};

	if (cstruct == NULL || priv == NULL)
		return;

	dbg("|%s:%d| Inside\n", __func__, __LINE__);

	t_cmdu_parse_tlvs(cstruct, tv, a_policy, 1);

	if (tv[0][0]) {
		topology_update_client_assoc_event(priv,
				(struct tlv_client_assoc_event *)tv[0][0]->data);
	} else {
		//here the TLv is not present so the
		//notification is from IEEE1905 running the algo
		dbg("|%s:%d| running the algo here\n", __func__, __LINE__);
		//if (!priv->algo_running)
		//	topologyd_algo_run(priv);
		uloop_timeout_set(&priv->topo_notify_timer, TOPO_NOTIFY_TIMER * 1000);
	}
}

uint16_t send_cmdu_query_resp(struct topologyd_private *priv, char *data,
		uint8_t *dst, uint8_t *src, uint16_t mid,
		uint16_t msg_type, char *ifname)
{
	struct blob_buf b = { 0 };
	char dst_addr[18] = { 0 };
	uint16_t msgid = 0;
	int ret = 0;
	uint32_t id;
	char dst_obj[256] = "ieee1905.al.";

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	blobmsg_add_u32(&b, "type", msg_type);



	blobmsg_add_u32(&b, "mid", mid);
	blobmsg_add_string(&b, "data", data);

	if ((hwaddr_is_zero(dst)) && (ifname != NULL)) {
		strncat(dst_obj, ifname, strlen(ifname));
		dbg("|%s:%d|object is :%s\n", __func__, __LINE__, dst_obj);
		if (ubus_lookup_id(priv->ctx, dst_obj, &id)) {
			dbg("[%s:%d] not present i1905 interface object", __func__, __LINE__);
			goto out;
		}

		hwaddr_ntoa(src, dst_addr);

		blobmsg_add_string(&b, "dst", dst_addr);
	}
	else {
		if (ubus_lookup_id(priv->ctx, I1905_OBJ, &id)) {
			dbg("[%s:%d] not present i1905", __func__, __LINE__);
			goto out;
		}

		if (dst == NULL)
			hwaddr_ntoa(priv->selfnode.hwaddr, dst_addr);
		else
			hwaddr_ntoa(dst, dst_addr);

		blobmsg_add_string(&b, "dst", dst_addr);
	}

	dbg("|%s:%d|cmdu:%u|dst:%s|mid:%u\n", __func__, __LINE__,
		msg_type, dst_addr, mid);

	ret = ubus_invoke(priv->ctx, id, "cmdu",
			b.head, send_cmdu_cb,
			&msgid,
			1000);
	if (ret) {
		dbg("[%s:%d] ubus call failed for |i1905 send|",
			__func__, __LINE__);
		goto out;
	}
	dbg("|%s:%d| msgid = %d\n", __func__, __LINE__, msgid);

	ret = msgid;
out:
	blob_buf_free(&b);

	return ret;
}


static void send_topology_buildcmdu_cb(struct ubus_request *req,
				       int type, struct blob_attr *msg)
{
	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	json_object *jroot = NULL;
	char *str;
	char *tlv_str = NULL;
	uint16_t msg_type = 0;
	int len = 0;
	struct topologyd_private *priv;
	uint8_t origin[6] = { 0 }, source[6] = { 0 };
	uint16_t msg_mid;
	char ifname[16] = { 0 };
	struct topo_resp_msg *msg_ptr = NULL;

	if (!msg)
		return;

	msg_ptr  = (struct topo_resp_msg *)req->priv;
	if (!msg_ptr)
		return;

	priv = msg_ptr->priv;
	msg_mid = msg_ptr->msg_mid;
	memcpy(origin, msg_ptr->origin, 6);
	memcpy(source, msg_ptr->source, 6);
	strncpy(ifname, msg_ptr->ifname, strlen(msg_ptr->ifname));

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		msg_type = _json_obj_get_int(jroot, "type");
		tlv_str = _json_obj_get_string(jroot, "data");
		if (tlv_str == NULL)
			return;
		len = strlen(tlv_str);
		if (len == 0) {
			err("|%s:%d| No data\n", __func__, __LINE__);
		}
		json_safe_put(jroot);
	}

	send_cmdu_query_resp(priv, tlv_str, origin, source, msg_mid,
		msg_type, ifname);

	if (tlv_str != NULL)
		free(tlv_str);
}

void topology_send_query_response(struct topologyd_private *priv,
	uint16_t msg_type, uint8_t *origin, uint8_t *source, uint16_t msg_mid,
	char *ifname)
{

	struct blob_buf b = { 0 };
	int ret = 0;
	uint32_t id;
	uint16_t resp_type;
	struct topo_resp_msg *msg_ptr;

	dbg("|%s:%d| Entry\n", __func__, __LINE__);

	if (origin && ifname)
		dbg("|%s:%d| ifname is %s origin is"MACFMT"\n",
				__func__, __LINE__, ifname, MAC2STR(origin));
	else
		return;

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	ret = get_resp_type(&resp_type, msg_type);
	if (ret != 0)
		goto out;

	msg_ptr = calloc(1, sizeof(struct topo_resp_msg));
	if (msg_ptr == NULL) {
		err("[%s:%d] NOMEM ", __func__, __LINE__);
		goto out;
	}

	msg_ptr->msg_mid = msg_mid;
	msg_ptr->priv = priv;
	memcpy(msg_ptr->origin, origin, 6);
	memcpy(msg_ptr->source, source, 6);
	strncpy(msg_ptr->ifname, ifname, strlen(ifname));

	blobmsg_add_u32(&b, "type", (uint32_t)resp_type);

	if (ubus_lookup_id(priv->ctx, I1905_OBJ, &id)) {
		dbg("[%s:%d] not present i1905", __func__, __LINE__);
		goto out_msg;
	}

	ret = ubus_invoke(priv->ctx, id, "buildcmdu",
			b.head, send_topology_buildcmdu_cb,
			msg_ptr,
			5000);
	if (ret)
		dbg("[%s:%d] ubus call failed for |i1905 buildcmdu|",
			__func__, __LINE__);

out_msg:
	free(msg_ptr);
out:
	blob_buf_free(&b);
	return;
}

static bool _update_ieee1905_node_info(struct topologyd_private *priv,
		struct node *old_node, struct node *new_node,
		int change_type)
{
	int found;
	int i, j;
	struct node *p = NULL;
	bool ret = false;

	dbg("%s() ===>\n", __func__);
	for (i = 0; i < old_node->nbr_num; i++) {
		found = 0;
		for (j = 0; j < new_node->nbr_num; j++) {
			if (!memcmp(old_node->nbr_list[i].nbr_macid,
					new_node->nbr_list[j].nbr_macid, 6)) {
				found = 1;
				break;
			}
		}

		if (found == 0) {
			ret = true;
			/* Lookup and delete the node as node
			 * not present in ieee1905 neighbors */
			p = node_lookup(priv->topo.node_htable,
					old_node->nbr_list[i].nbr_macid);
			if (p)
				uloop_timeout_set(&p->validity_timer, 1);
		}
	}
	return ret;
}

void update_ieee1905_node_info(struct topologyd_private *priv,
		struct node *old_node, struct node *new_node)
{
	bool is_change = false;
	dbg("Inside %s..\n", __func__);

	if (old_node == NULL || new_node == NULL || priv == NULL)
		return;

	dbg("The dest num [%d] src num [%d]", old_node->nbr_num,
		new_node->nbr_num);

	is_change = _update_ieee1905_node_info(priv, old_node, new_node, NBR_NODE_DEL);

	if (old_node->nbr_num < new_node->nbr_num || is_change)
		uloop_timeout_set(&priv->refresh_timer, 5);
}

/* This function is used to update the bridge node info in case of R2
 * when higher layer is not present
 */
void host_update_bridgenode(struct topologyd_private *priv,
		struct node *n, int intf_num, uint8_t *tif)
{
	struct host_node *node;
	char mac_str[18] = { 0 };
	uint8_t if_macaddr[6];

	dbg("Inside %s..\n", __func__);
	/*Here we check if the host is present in the dhcp then add that to node list*/
	if (host_check_dhcpv4_info(tif, intf_num, if_macaddr)) {
		struct host_node *iface_brnode = NULL;
		int rv = 0;

		iface_brnode = host_node_lookup(priv->host.node_htable, if_macaddr);
		dbg("%s %d bridge Node mac = ["MACFMT"]\n", __func__, __LINE__, MAC2STR(if_macaddr));
		if (!iface_brnode) {
			iface_brnode = host_topo_node_add(priv, 0, if_macaddr, 0);
			if (!iface_brnode)
				err("Failed to add node in the hosts lookup\n");
		}

		node = host_node_lookup(priv->host.node_htable, n->hwaddr);
		if (node) {
			rv = memcmp(node->hwaddr, iface_brnode->hwaddr, 6);
			if (rv != 0) {
				dbg("%s %d we found al node \n", __func__, __LINE__);
				node->ignore = true;
				iface_brnode->ignore = false;
				memcpy(iface_brnode->al_node, node->hwaddr, HW_ADDR_LENGTH);
				iface_brnode->intf_type = node->intf_type;
				hwaddr_ntoa(iface_brnode->hwaddr, mac_str);
				config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type",
					(iface_brnode->intf_type == HOST_TYPE_WIFI ? "wifi" : "eth"),
					priv->host_global_cfg.persistent);
				iface_brnode->is_device_mac = 1;
				//Here we need a variable so that the node value is not changed
				//periodically
				node->is_copy = 1;
				//Here we need to copy the data of the br-lan
				//to that of the al-mac node
				host_copy_node_data(iface_brnode, node, priv);
				if (node->is_event == 0) {
					host_send_client_event(priv, node, 1);
					node->is_event = 1;
				}
			} else {
				node->is_copy = 0;
			}
		}
	}
}
